/*****************************************************************************\
 *                                                                           *
 * Generator.h                                                               *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#ifndef GENERATOR_H
#define GENERATOR_H

#include <Essential/Hashtable.h>
#include "Xrtti/XrttiParsed.h"
#include "private/Configuration.h"

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


class Generator;
class GeneratorArgument;
class GeneratorArray;
class GeneratorArrayOrPointer;
class GeneratorBase;
class GeneratorClass;
class GeneratorConstructor;
class GeneratorConstructorSignature;
class GeneratorContext;
class GeneratorDestructor;
class GeneratorDestructorSignature;
class GeneratorEnumeration;
class GeneratorEnumerationValue;
class GeneratorField;
class GeneratorImplementation;
class GeneratorMethod;
class GeneratorMethodSignature;
class GeneratorNamespace;
class GeneratorObject;
class GeneratorPointer;
class GeneratorStructure;
class GeneratorStruct;
class GeneratorType;
class GeneratorTypeEnumeration;
class GeneratorTypeFunction;
class GeneratorTypeStructure;
class GeneratorTypeUnion;
class GeneratorUnion;


class GeneratorObject
{
public:

    GeneratorObject()
        : numberM(nextNumberG++)
    {
    }

    virtual ~GeneratorObject()
    {
    }

    U32 GetNumber()
    {
        return numberM;
    }

    // Implemented by subclasses to define any XrttiAccess output which needs
    // to happen in the XrttiAccess class definition.  Subclasses should
    // always call their superclass EmitXrttiAccess() method first.
    virtual void EmitXrttiAccess(OutputStream &out)
    {
    }

    // Implemented by subclasses to emit the definition of static fields
    // defined in XrttiAccess
    virtual void EmitXrttiAccessStaticDefinitions(OutputStream &out)
    {
    }

    // Emits the declarations which need to happen in the _ class
    void EmitDeclaration(OutputStream &out);

    // Emits the definitions of the declarations from EmitDeclarations
    void EmitDefinition(OutputStream &out);

protected:

    // Implemented by subclasses to return the Compiled type to create at
    // runtime for this object
    virtual const char *GetCompiledTypeName() = 0;

    // Implemented by subclasses to emit any extra declarations aside from
    // the declaration of this object itself.  Subclasses should always
    // call their superclass EmitExtraDeclarations() method first.
    virtual void EmitExtraDeclarations(OutputStream &out)
    {
    }

    // Implemented by subclasses to emit any extra definitions aside from
    // the definition of this object itself.  Subclasses should always
    // call their superclass EmitExtraDefinitions() method first.
    virtual void EmitExtraDefinitions(OutputStream &out)
    {
    }

    // Implemented by subclasses to emit any arguments for this specific class
    // in the constructor of the Compiled representation of this object.
    // Subclasses should always call their superclass
    // EmitDefinitionArguments() method first.
    virtual void EmitDefinitionArguments(OutputStream &out) = 0;

private:

    U32 numberM;

    static U32 nextNumberG;
};


class GeneratorContext : public GeneratorObject
{
public:

    GeneratorContext(Generator &generator, const Context &context);

    bool IsAccessible();

    String GetTypeName();

    void EmitTypedefs(OutputStream &out);

protected:

    virtual void EmitNeededTypedefs(OutputStream &out) = 0;

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    bool typedefsEmittedM;

    const Context &contextM;

    GeneratorContext *pContextM;
};


class GeneratorNamespace : public GeneratorContext
{
public:

    GeneratorNamespace(Generator &generator, const Context &context)
        : GeneratorContext(generator, context)
    {
    }

protected:

    virtual void EmitNeededTypedefs(OutputStream &out)
    {
    }

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledNamespace";
    }
};


class GeneratorStructure : public GeneratorContext
{
public:

    GeneratorStructure(Generator &generator, const Structure &structure);

    virtual ~GeneratorStructure();

    virtual void EmitXrttiAccess(OutputStream &out);

    virtual void EmitXrttiAccessStaticDefinitions(OutputStream &out);

protected:

    virtual void EmitNeededTypedefs(OutputStream &out);

    bool CanInvokeDefaultConstructor();

    bool CanInvokeDefaultDestructor();

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    bool rttiM;

    const Structure &structureM;

    GeneratorBase **pBasesM;

    GeneratorStructure **pFriendsM;

    GeneratorField **pFieldsM;

    GeneratorConstructor **pConstructorsM;

    GeneratorDestructor *pDestructorM;
};


class GeneratorBase : public GeneratorObject
{
public:

    GeneratorBase(Generator &generator, const Base &base,
                  GeneratorStructure *pSubstructure);

    virtual void EmitXrttiAccess(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledBase";
    }

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    bool CanEmitCast();

    const Base &baseM;

    GeneratorStructure *pSubstructureM;

    GeneratorStructure *pStructureM;
};


class GeneratorUnion : public GeneratorStructure
{
public:

    GeneratorUnion(Generator &generator, const Union &unionRef)
        : GeneratorStructure(generator, (const Structure &) unionRef)
    {
    }

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledUnion";
    }
};


class GeneratorStruct : public GeneratorStructure
{
public:

    GeneratorStruct(Generator &generator, const Struct &structRef);

    virtual ~GeneratorStruct();

    virtual void EmitXrttiAccess(OutputStream &out);

protected:

    virtual void EmitNeededTypedefs(OutputStream &out);

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledStruct";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const Struct &structM;

    GeneratorMethod **pMethodsM;
};


class GeneratorClass : public GeneratorStruct
{
public:

    GeneratorClass(Generator &generator, const Class &classRef)
        : GeneratorStruct(generator, (const Struct &) classRef)
    {
    }

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledClass";
    }
};


class GeneratorMember : public GeneratorObject
{
public:
    
    GeneratorMember(Generator &generator, const Member &member);

    bool IsAccessible();

    const char *GetName()
    {
        return memberM.GetName();
    }

protected:

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const Member &memberM;

    GeneratorContext *pContextM;
};


class GeneratorField : public GeneratorMember
{
public:

    GeneratorField(Generator &generator, const Field &field);

    void EmitTypedefs(OutputStream &out);

    virtual void EmitXrttiAccess(OutputStream &out);

    virtual void EmitXrttiAccessStaticDefinitions(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledField";
    }

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    bool CanEmitOffsetof();

    bool CanEmitGet();

    void EmitGet(OutputStream &out);

    const Field &fieldM;

    GeneratorType *pTypeM;
};


class GeneratorArgument : public GeneratorObject
{
public:

    GeneratorArgument(Generator &generator, const Argument &argument);

    const char *GetTypeConstyThing();

    U32 GetTypeNumber();

    bool CanEmitType();

    void EmitTypedefs(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledArgument";
    }

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const Argument &argumentM;

    GeneratorType *pTypeM;
};


class GeneratorDestructorSignature : public GeneratorObject
{
public:

    GeneratorDestructorSignature(Generator &generator,
                                 const DestructorSignature &destructorSig);

    virtual ~GeneratorDestructorSignature();

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledDestructorSignature";
    }

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const DestructorSignature &destructorSignatureM;
    
    GeneratorType **pThrowsM;
};


class GeneratorConstructorSignature : public GeneratorDestructorSignature
{
public:

    GeneratorConstructorSignature(Generator &generator,
                                  const ConstructorSignature &sig);

    virtual ~GeneratorConstructorSignature();

    U32 GetArgumentCount()
    {
        return constructorSignatureM.GetArgumentCount();
    }

    U32 GetArgumentTypeNumber(U32 index)
    {
        return pArgumentsM[index]->GetTypeNumber();
    }

    bool CanEmitArguments();

    void EmitInvokeArguments(OutputStream &out);

    void EmitTypedefs(OutputStream &out);

    void EmitArgumentTypeList(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledConstructorSignature";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const ConstructorSignature &constructorSignatureM;

    GeneratorArgument **pArgumentsM;
};


class GeneratorMethodSignature : public GeneratorConstructorSignature
{
public:
    
    GeneratorMethodSignature(Generator &generator, const MethodSignature &sig);
    
    U32 GetReturnTypeNumber();

    bool CanEmitReturnTypeAndArguments();

    void EmitTypedefs(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledMethodSignature";
    }

    virtual void EmitDefinitionArguments(OutputStream &out);
    
private:

    const MethodSignature &methodSignatureM;

    GeneratorType *pReturnTypeM;
};


class GeneratorDestructor : public GeneratorMember
{
public:

    GeneratorDestructor(Generator &generator, const Destructor &destructor)
        : GeneratorMember(generator, (const Member &) destructor),
          destructorM(destructor), 
          destructorSignatureM(generator, destructor.GetSignature())
    {
    }

    bool IsAccessible();

    bool CanInvoke();

    virtual void EmitXrttiAccess(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledDestructor";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:
    
    const Destructor &destructorM;

    GeneratorDestructorSignature destructorSignatureM;
};


class GeneratorConstructor : public GeneratorMember
{
public:

    GeneratorConstructor(Generator &generator, const Constructor &constructor)
        : GeneratorMember(generator, (const Member &) constructor),
          constructorM(constructor), 
          constructorSignatureM(generator, constructor.GetSignature())
    {
    }

    bool IsDefault()
    {
        return !constructorM.GetSignature().GetArgumentCount();
    }

    bool IsAccessible();

    bool CanInvoke();

    virtual void EmitTypedefs(OutputStream &out);

    void EmitXrttiAccess(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledConstructor";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const Constructor &constructorM;

    GeneratorConstructorSignature constructorSignatureM;
};


class GeneratorMethod : public GeneratorMember
{
public:

    GeneratorMethod(Generator &generator, const Method &method);

    bool CanInvoke();

    virtual void EmitTypedefs(OutputStream &out);

    void EmitXrttiAccess(OutputStream &out);

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledMethod";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const Method &methodM;

    GeneratorMethodSignature methodSignatureM;
};


class GeneratorType : public GeneratorObject
{
public:

    GeneratorType(Generator &generator, const Type &type);

    virtual ~GeneratorType();

    virtual bool CanEmitTypedef()
    {
        return true;
    }

    virtual void EmitTypedef(OutputStream &out);

protected:

    virtual void EmitPrerequisiteTypedefs(OutputStream &out)
    {
    }

    virtual void EmitTypedefFunctionArguments(OutputStream &out)
    {
    }

    virtual String GetTypedefTypeName();

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledType";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const char *GetBaseTypeName();

    String GetCppTypeName();

    bool typedefEmittedM;

    const Type &typeM;

    GeneratorArrayOrPointer **pArrayOrPointersM;
};


class GeneratorArrayOrPointer : public GeneratorObject
{
public:

    GeneratorArrayOrPointer()
    {
    }

    virtual ArrayOrPointer::Type GetType() = 0;
};


class GeneratorArray : public GeneratorArrayOrPointer
{
public:

    GeneratorArray(const Array &array)
        : arrayM(array)
    {
    }

    virtual ArrayOrPointer::Type GetType()
    {
        return ArrayOrPointer::Type_Array;
    }

protected:
    
    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledArray";
    }
    
    virtual void EmitDefinitionArguments(OutputStream &out);

private:
    
    const Array &arrayM;
};


class GeneratorPointer : public GeneratorArrayOrPointer
{
public:

    GeneratorPointer(const Pointer &pointer)
        : pointerM(pointer)
    {
    }

    virtual ArrayOrPointer::Type GetType()
    {
        return ArrayOrPointer::Type_Pointer;
    }

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledPointer";
    }
    
    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const Pointer &pointerM;
};


class GeneratorEnumeration : public GeneratorObject
{
public:

    GeneratorEnumeration(Generator &generator, const Enumeration &enumeration);

    virtual ~GeneratorEnumeration();

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledEnumeration";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const Enumeration &enumerationM;

    GeneratorContext *pContextM;

    GeneratorEnumerationValue **pValuesM;
};


class GeneratorEnumerationValue : public GeneratorObject
{
public:

    GeneratorEnumerationValue(const EnumerationValue &value)
        : valueM(value)
    {
    }

protected:

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledEnumerationValue";
    }

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const EnumerationValue &valueM;
};


class GeneratorTypeEnumeration : public GeneratorType
{
public:

    GeneratorTypeEnumeration(Generator &generator,
                             const TypeEnumeration &typeEnumeration);

    virtual bool CanEmitTypedef();

protected:

    virtual String GetTypedefTypeName();

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledTypeEnumeration";
    }

    virtual void EmitDefinitionArguments(OutputStream &out);

private:
    
    const TypeEnumeration &typeEnumerationM;

    GeneratorEnumeration *pEnumerationM;
};


class GeneratorTypeFunction : public GeneratorType
{
public:

    GeneratorTypeFunction(Generator &generator, 
                          const TypeFunction &typeFunction);

    virtual bool CanEmitTypedef();

protected:

    virtual void EmitPrerequisiteTypedefs(OutputStream &out);

    virtual void EmitTypedefFunctionArguments(OutputStream &out);

    virtual String GetTypedefTypeName();

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledTypeFunction";
    }

    virtual void EmitExtraDeclarations(OutputStream &out);

    virtual void EmitExtraDefinitions(OutputStream &out);

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const TypeFunction &typeFunctionM;
    
    GeneratorMethodSignature *pMethodSignatureM;
};


class GeneratorTypeStructure : public GeneratorType
{
public:

    GeneratorTypeStructure(Generator &generator,
                           const TypeStructure &typeStructure);

    virtual bool CanEmitTypedef();

protected:

    virtual String GetTypedefTypeName();

    virtual const char *GetCompiledTypeName()
    {
        return "Xrtti::CompiledTypeStructure";
    }

    virtual void EmitDefinitionArguments(OutputStream &out);

private:

    const TypeStructure &typeStructureM;
    
    GeneratorStructure *pStructureM;
};


class Generator
{
public:

    Generator(Configuration &config, const ContextSet &contextSet);
    ~Generator();

    bool Generate();

    const Configuration &GetConfiguration()
    {
        return configM;
    }

    GeneratorContext *GetGeneratorContext(const Context *pContext);
    GeneratorType *GetGeneratorType(const Type *pType);
    GeneratorEnumeration *GetGeneratorEnumeration(const Enumeration *pEnum);

    void RegisterGeneratorContext(const Context *pContext,
                                  GeneratorContext *pGeneratorContext);
    void RegisterGeneratorType(const Type *pType,
                               GeneratorType *pGeneratorType);
    void RegisterGeneratorEnumeration(const Enumeration *pEnumeration,
                                      GeneratorEnumeration *pGeneratorEnum);

    static void EmitBooleanArgument(OutputStream &out, bool val, bool comma);
    static void EmitStringArgument(OutputStream &out, const char *str,
                                   bool comma);
    static void EmitU32Argument(OutputStream &out, U32 number, bool comma);
    static void EmitAccessTypeArgument(OutputStream &out, 
                                       AccessType accessType, bool comma);
    static void EmitArrayArgument(OutputStream &out, U32 count,
                                  const char *name, U32 number, bool comma);

    static bool IsAccessible(const Context *pContext);
    static bool IsAccessible(const Member *pMember);
    static bool IsPod(const Context *pContext);

    static String GetTypeName(const Context *pContext);

private:

    static bool IsXrttiAccessFriend(const Structure *pStructure);

    void EmitHeader(OutputStream &out);
    void EmitXrttiAccess(OutputStream &out);
    void EmitXrttiAccessStaticDefinitions(OutputStream &out);
    void EmitTypedefs(OutputStream &out);
    void EmitDeclarations(OutputStream &out);
    void EmitDefinitions(OutputStream &out);
    void EmitRegister(OutputStream &out);

    Configuration &configM;

    Hashtable<const Context *, GeneratorContext *> htGeneratorContextsM;

    Hashtable<const Type *, GeneratorType *> htGeneratorTypesM;

    Hashtable<const Enumeration *, GeneratorEnumeration *> 
        htGeneratorEnumerationsM;
};


}; // namespace Xrtti

#endif // GENERATOR_H
