/*****************************************************************************\
 *                                                                           *
 * Parser.h                                                                  *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
 * Defines a parser class which handles parsing of gccxml output into a      *
 * set of C++ objects hashed by id.                                          *
 *                                                                           *
\*****************************************************************************/

#ifndef PARSER_H
#define PARSER_H

#include <Essential/InputStream.h>
#include <Essential/RefHashtable.h>
#include <Essential/RefVector.h>
#include <Essential/String.h>

#include <expat.h>

using namespace Essential;

// ---------------------------------------------------------------------------
// Parser
// ---------------------------------------------------------------------------
class Parser
{
public:

    enum AccessType
    {
        AccessTypePublic,
        AccessTypeProtected,
        AccessTypePrivate
    };

    enum SubType
    {
        SubTypeFundamental,
        SubTypePointer,
        SubTypeReference,
        SubTypeCvQualified,
        SubTypeFunction
    };

    // -----------------------------------------------------------------------
    // Parser::Attribute
    // -----------------------------------------------------------------------
    class Attribute : public RefCounted
    {
    public:

        Attribute(char *pName, char *pValue);

        String GetName() const
        {
            return nameM;
        }

        String GetValue() const
        {
            return valueM;
        }
    private:

        const String nameM;
        const String valueM;
    };

    // -----------------------------------------------------------------------
    // Parser::Element
    // -----------------------------------------------------------------------
    class Element : public RefCounted
    {
    public:

        Element(const Parser &parser, char *pElementName, char **pAttributes);

        String GetElementName() const
        {
            return elementNameM;
        }

        bool HasAttribute(const String &name) const
        {
            U32 count = this->GetAttributeCount();
            
            for (U32 i = 0; i < count; i++) {
                ref<Attribute> pAttribute = this->GetAttribute(i);
                if (pAttribute->GetName() == name) {
                    return true;
                }
            }

            return false;
        }

        ref<Attribute> GetAttribute(const String &name) const
        {
            U32 count = this->GetAttributeCount();
            
            for (U32 i = 0; i < count; i++) {
                ref<Attribute> pAttribute = this->GetAttribute(i);
                if (pAttribute->GetName() == name) {
                    return pAttribute;
                }
            }

            return NULL;
        }

        String GetAttributeValue(const String &name) const
        {
            ref<Attribute> pAttribute = this->GetAttribute(name);

            if (pAttribute == NULL) {
                return "";
            }

            return pAttribute->GetValue();
        }

        // This is the line number that the element was declared on
        U32 GetLineNumber() const
        {
            return lineNumberM;
        }

        ref<Element> GetSubElement(const String &name) const
        {
            U32 count = this->GetSubElementCount();

            for (U32 i = 0; i < count; i++) {
                ref<Element> pElement = this->GetSubElement(i);
                if (pElement->GetElementName() == name) {
                    return pElement;
                }
            }

            return NULL;
        }

        U32 GetAttributeCount() const
        {
            return attributesM.Count();
        }

        ref<Attribute> GetAttribute(U32 index) const
        {
            return attributesM[index];
        }

        U32 GetSubElementCount() const
        {
            return subElementsM.Count();
        }
        
        ref<Element> GetSubElement(U32 index) const
        {
            return subElementsM[index];
        }

    private:

        friend class Parser;

        const String elementNameM;
        RefVector<Attribute> attributesM;
        RefVector<Element> subElementsM;
        U32 lineNumberM;
    };

    // -----------------------------------------------------------------------
    // Parser::ElementEnumeration
    //
    // This creates an enumeration of all elements which have been parsed
    // of the given elementName
    // -----------------------------------------------------------------------
    class ElementEnumeration : public RefEnumeration<Element>
    {
    public:

        ElementEnumeration(const Parser &parser)
            : elementEnumeration(parser.htElementsM) { }

        virtual bool HasElements() const
        {
            return elementEnumeration.HasElements();
        }

        virtual ref<Element> GetNextElement()
        {
            return elementEnumeration.GetNextElement();
        }

    private:

        RefHashtable__ValueEnumeration<String, Element> elementEnumeration;
    };

    // Creates a Parser object
    Parser();

    // Parses the inputstream, which is an XML document as output by gccxml;
    // uses the given configuration settings
    void Parse(InputStream &instream);

    // Get and return an element by its id
    ref<Element> LookupElement(const String &id) const
    {
        return htElementsM.Get(id);
    }

private:

    // XML parsing methods
    static void ExpatStartElementHandler(void *pUserData,
                                         const XML_Char *pName,
                                         const XML_Char **pAttributes);
    static void ExpatEndElementHandler(void *pUserData,
                                       const XML_Char *pName);
    void StartElement(char *pElementName, char **pAttributes);
    void EndElement(char *pElementName);
    void XmlError(const char *pFormat, ...) const;
    U32 GetCurrentParserLineNumber() const;

    // XML parsing fields
    bool parseErrorM;
    XML_Parser xmlParserM;

    // The content which was parsed
    RefHashtable<String, Element> htElementsM;

    // The current stack of Elements
    RefVector<Element> vCurrentElementsM;
};

#endif // PARSER_H
