/*****************************************************************************\
 *                                                                           *
 * TestClasses.h                                                             *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#ifndef TEST_CLASSES_H
#define TEST_CLASSES_H

#include <Essential/DataTypes.h>
#include <Essential/File.h>
#include <Essential/Log.h>
#include <Essential/String.h>
#include <Essential/Util.h>

using namespace Essential;

// Convenience for setting a value to a random number
template<typename T>
inline void Rand(T &t)
{
    t = Util::RandomNumber();
}

// Convenience for setting a value to a random real number
template<typename T>
inline void RandReal(T &t)
{
    t = ((T) Util::RandomNumber()) / ((T) Util::RandomNumber());

    if (Util::RandomNumber() % 2) {
        t = -t;
    }
}

// Convenience for setting a value to a random number, with a maximum
template<typename T>
inline void Rand(T &t, T max)
{
    t = Util::RandomNumber() % max;
}

// Convenience for setting an inline max-terminated array to random real 
// numbers
template<typename T>
inline void RandRealArray(T *pArray, U32 maxCount)
{
    for (U32 i = 0; i < maxCount; i++) {
        RandReal<T>(pArray[i]);
    }
}

// Convenience for setting a non-inline max-terminated array to random real 
// numbers
template<typename T>
inline void CreateRandRealArray(T *&pArray, U32 maxCount)
{
    pArray = new T[maxCount];

    for (U32 i = 0; i < maxCount; i++) {
        RandReal<T>(pArray[i]);
    }
}

// Convenience for setting an inline count-terminated array to random numbers
template<typename T, typename C>
inline void RandArray(T *pArray, C &count, U32 maxCount)
{
    count = Util::RandomNumber() % (maxCount + 1);
    
    for (U32 i = 0; i < (U32) count; i++) {
        Rand<T>(pArray[i]);
    }
}

// Convenience for setting an inline count-terminated array to random real 
// numbers
template<typename T, typename C>
inline void RandRealArray(T *pArray, C &count, U32 maxCount)
{
    count = Util::RandomNumber() % (maxCount + 1);
    
    for (U32 i = 0; i < (U32) count; i++) {
        RandReal<T>(pArray[i]);
    }
}

// Convenience for setting a non-inline count-terminated array to random 
// numbers
template<typename T, typename C>
inline void CreateRandArray(T *&pArray, C &count, U32 maxCount)
{
    count = Util::RandomNumber() % (maxCount + 1);

    pArray = new T[count];
    
    for (U32 i = 0; i < (U32) count; i++) {
        Rand<T>(pArray[i]);
    }
}

// Convenience for setting a non-inline count-terminated array to random real 
// numbers
template<typename T, typename C>
inline void CreateRandRealArray(T *&pArray, C &count, U32 maxCount)
{
    count = Util::RandomNumber() % (maxCount + 1);

    pArray = new T[count];

    for (U32 i = 0; i < (U32) count; i++) {
        RandReal<T>(pArray[i]);
    }
}

// Convenience for setting an inline zero-terminated array to random numbers
template<typename T>
inline void RandArray(T *pArray, U32 maxCount)
{
    U32 count = Util::RandomNumber() % (maxCount + 1);
    
    for (U32 i = 0; i < count; i++) {
        Rand<T>(pArray[i]);
    }

    if (count < maxCount) {
        pArray[count] = 0;
    }
}

// Convenience for setting a non-inline zero-terminated array to random numbers
template<typename T>
inline void CreateRandArray(T *&pArray, U32 maxCount)
{
    U32 count = Util::RandomNumber() % maxCount;
    
    pArray = new T[count + 1];
    
    for (U32 i = 0; i < (U32) count; i++) {
        Rand<T>(pArray[i]);
    }

    pArray[count] = 0;
}

inline void LogPrefix(OutputStream &out, const String &prefix, 
                      const char *format, ...)
{
    out.WriteA(prefix);

    va_list valist;

    va_start(valist, format);

    out.WriteA(String::FormatV(format, valist));

    va_end(valist);
}

class TestClass
{
public:

    static TestClass *Create(bool fixedId, U32 id);

    inline TestClass()
        : isLastM(false)
    {
    }
    
    inline TestClass(bool isLast)
        : isLastM(isLast)
    {
    }
    
    virtual ~TestClass() { }

    void SetIsLast(bool isLast)
    {
        isLastM = isLast;
    }

    bool IsLast()
    {
        return isLastM;
    }

    virtual void Randomize() { }

    virtual const char *GetIdentity() const { return "TestClass"; }

private:

    bool isLastM;
};

// The purpose of these is to exercise every permutation of serialization

class BasicTypes : public TestClass
{
public:

    // The actual value of fRandomize does not matter
    virtual void Randomize()
    {
        Rand<S8>(s8);
        Rand<U8>(u8);
        Rand<S16>(s16);
        Rand<U16>(u16);
        Rand<S32>(s32);
        Rand<U32>(u32);
        Rand<S64>(s64);
        Rand<U64>(u64);
        RandReal<float>(fl);
        RandReal<double>(dl);
    }

    virtual const char *GetIdentity() const
    {
        return "BasicTypes";
    }

private:
    
    S8 s8;

    U8 u8;

    S16 s16;

    U16 u16;

    S32 s32;

    U32 u32;

    S64 s64;

    U64 u64;

    float fl;

    double dl;
};

class BasicTypeArrays : public TestClass
{
public:
    
    // The actual value of fRandomize does not matter
    virtual void Randomize()
    {
        RandArray<S8>(s8, Util__CountOf(s8));
        RandArray<U8>(u8, Util__CountOf(u8));
        RandArray<S16>(s16, Util__CountOf(s16));
        RandArray<U16>(u16, Util__CountOf(u16));
        RandArray<S32>(s32, Util__CountOf(s32));
        RandArray<U32>(u32, Util__CountOf(u32));
        RandArray<S64>(s64, Util__CountOf(s64));
        RandArray<U64>(u64, Util__CountOf(u64));
        RandRealArray<float>(fl, Util__CountOf(fl));
        RandRealArray<double>(dl, Util__CountOf(dl));
    }

    virtual const char *GetIdentity() const
    {
        return "BasicTypeArrays";
    }

private:

    S8 s8[8];

    U8 u8[8];

    S16 s16[16];

    U16 u16[16];

    S32 s32[32];

    U32 u32[32];

    S64 s64[64];

    U64 u64[64];

    float fl[10];

    double dl[10];
};

class BasicTypeArraysWithCount : public TestClass
{
public:

    inline BasicTypeArraysWithCount()
        : s8_CountM(0), u8_CountM(0), s16_CountM(0), u16_CountM(0),
          s32_CountM(0), u32_CountM(0), s64_CountM(0), u64_CountM(0),
          fl_CountM(0), dl_CountM(0)
    {
    }

    virtual void Randomize()
    {
        RandArray<S8, S8>(s8, s8_CountM, Util__CountOf(s8));
        RandArray<U8, U8>(u8, u8_CountM, Util__CountOf(u8));
        RandArray<S16, S16>(s16, s16_CountM, Util__CountOf(s16));
        RandArray<U16, U16>(u16, u16_CountM, Util__CountOf(u16));
        RandArray<S32, S32>(s32, s32_CountM, Util__CountOf(s32));
        RandArray<U32, U32>(u32, u32_CountM, Util__CountOf(u32));
        RandArray<S64, S32>(s64, s64_CountM, Util__CountOf(s64));
        RandArray<U64, U32>(u64, u64_CountM, Util__CountOf(u64));
        RandRealArray<float, U32>(fl, fl_CountM, Util__CountOf(fl));
        RandRealArray<double, U32>(dl, dl_CountM, Util__CountOf(dl));
    }

    virtual const char *GetIdentity() const
    {
        return "BasicTypeArraysWithCount";
    }

    void bleh(int a, char *c = "hello")
    {
    }

private:

    S8 s8_CountM;
    S8 s8[8];

    U8 u8_CountM; 
    U8 u8[8];

    S16 s16_CountM;
    S16 s16[16];

    U16 u16_CountM;
    U16 u16[16];

    S32 s32_CountM;
    S32 s32[32];

    U32 u32_CountM;
    U32 u32[32];

    S32 s64_CountM;
    S64 s64[64];

    U32 u64_CountM;
    U64 u64[64];

    U32 fl_CountM;
    float fl[10];

    U32 dl_CountM;
    double dl[10];
};

class BasicTypePointersWithCount : public TestClass
{
public:

    inline BasicTypePointersWithCount()
        : pS8_CountM(0), pS8(NULL), pU8_CountM(0), pU8(NULL), 
          pS16_CountM(0), pS16(NULL), pU16_CountM(0), pU16(NULL), 
          pS32_CountM(0), pS32(NULL), pU32_CountM(0), pU32(NULL), 
          pS64_CountM(0), pS64(NULL), pU64_CountM(0), pU64(NULL), 
          pFl_CountM(0), pFl(NULL), pDl_CountM(0), pDl(NULL)
    {
    }

    virtual ~BasicTypePointersWithCount()
    {
        delete [] pS8;
        delete [] pU8;
        delete [] pS16;
        delete [] pU16;
        delete [] pS32;
        delete [] pU32;
        delete [] pS64;
        delete [] pU64;
        delete [] pFl;
        delete [] pDl;
    }

    virtual void Randomize()
    {
        CreateRandArray<S8, S8>(pS8, pS8_CountM, 100);
        CreateRandArray<U8, U8>(pU8, pU8_CountM, 200);
        CreateRandArray<S16, S16>(pS16, pS16_CountM, 50);
        CreateRandArray<U16, U16>(pU16, pU16_CountM, 75);
        CreateRandArray<S32, S32>(pS32, pS32_CountM, 123);
        CreateRandArray<U32, U32>(pU32, pU32_CountM, 321);
        CreateRandArray<S64, S32>(pS64, pS64_CountM, 1);
        CreateRandArray<U64, U32>(pU64, pU64_CountM, 6);
        CreateRandRealArray<float, U32>(pFl, pFl_CountM, 9);
        CreateRandRealArray<double, U32>(pDl, pDl_CountM, 333);
    }

    virtual const char *GetIdentity() const
    {
        return "BasicTypePointersWithCount";
    }

private:

    S8 pS8_CountM;
    S8 *pS8;

    U8 pU8_CountM; 
    U8 *pU8;

    S16 pS16_CountM;
    S16 *pS16;

    U16 pU16_CountM;
    U16 *pU16;

    S32 pS32_CountM;
    S32 *pS32;

    U32 pU32_CountM;
    U32 *pU32;

    S32 pS64_CountM;
    S64 *pS64;

    U32 pU64_CountM;
    U64 *pU64;

    U32 pFl_CountM;
    float *pFl;

    U32 pDl_CountM;
    double *pDl;
};

class BasicTypePointersWithMax : public TestClass
{
public:

    inline BasicTypePointersWithMax()
        : pS8(NULL), pU8(NULL), pS16(NULL), pU16(NULL), pS32(NULL), pU32(NULL),
          pS64(NULL), pU64(NULL), pFl(NULL), pDl(NULL)
    {
    }

    virtual ~BasicTypePointersWithMax()
    {
        delete [] pS8;
        delete [] pU8;
        delete [] pS16;
        delete [] pU16;
        delete [] pS32;
        delete [] pU32;
        delete [] pS64;
        delete [] pU64;
        delete [] pFl;
        delete [] pDl;
    }

    virtual void Randomize()
    {
        CreateRandArray<S8>(pS8, pS8_Max);
        CreateRandArray<U8>(pU8, pU8_Max);
        CreateRandArray<S16>(pS16, pS16_Max);
        CreateRandArray<U16>(pU16, pU16_Max);
        CreateRandArray<S32>(pS32, pS32_Max);
        CreateRandArray<U32>(pU32, pU32_Max);
        CreateRandArray<S64>(pS64, pS64_Max);
        CreateRandArray<U64>(pU64, pU64_Max);
        CreateRandRealArray<float>(pFl, pFl_Max);
        CreateRandRealArray<double>(pDl, pDl_Max);
    }

    virtual const char *GetIdentity() const
    {
        return "BasicTypePointersWithMax";
    }

private:

    static const int pS8_Max = 8;
    S8 *pS8;

    static const int pU8_Max = 8;
    U8 *pU8;

    static const int pS16_Max = 16;
    S16 *pS16;

    static const int pU16_Max = 16;
    U16 *pU16;

    static const int pS32_Max = 32;
    S32 *pS32;

    static const int pU32_Max = 32;
    U32 *pU32;

    static const int pS64_Max = 64;
    S64 *pS64;

    static const int pU64_Max = 64;
    U64 *pU64;

    static const int pFl_Max = 32;
    float *pFl;

    static const int pDl_Max = 64;
    double *pDl;
};

class BasicTypePointersWithCountAndMax : public TestClass
{
public:

    inline BasicTypePointersWithCountAndMax()
        : pS8_CountM(0), pS8(NULL), pU8_CountM(0), pU8(NULL), 
          pS16_CountM(0), pS16(NULL), pU16_CountM(0), pU16(NULL), 
          pS32_CountM(0), pS32(NULL), pU32_CountM(0), pU32(NULL), 
          pS64_CountM(0), pS64(NULL), pU64_CountM(0), pU64(NULL), 
          pFl_CountM(0), pFl(NULL), pDl_CountM(0), pDl(NULL)
    {
    }

    virtual ~BasicTypePointersWithCountAndMax()
    {
        delete [] pS8;
        delete [] pU8;
        delete [] pS16;
        delete [] pU16;
        delete [] pS32;
        delete [] pU32;
        delete [] pS64;
        delete [] pU64;
        delete [] pFl;
        delete [] pDl;
    }

    virtual void Randomize()
    {
        CreateRandArray<S8, S8>(pS8, pS8_CountM, pS8_Max);
        CreateRandArray<U8, U8>(pU8, pU8_CountM, pU8_Max);
        CreateRandArray<S16, S16>(pS16, pS16_CountM, pS16_Max);
        CreateRandArray<U16, U16>(pU16, pU16_CountM, pU16_Max);
        CreateRandArray<S32, S32>(pS32, pS32_CountM, pS32_Max);
        CreateRandArray<U32, U32>(pU32, pU32_CountM, pU32_Max);
        CreateRandArray<S64, S32>(pS64, pS64_CountM, pS64_Max);
        CreateRandArray<U64, U32>(pU64, pU64_CountM, pU64_Max);
        CreateRandRealArray<float, U32>(pFl, pFl_CountM, pFl_Max);
        CreateRandRealArray<double, U32>(pDl, pDl_CountM, pDl_Max);
    }

    virtual const char *GetIdentity() const
    {
        return "BasicTypePointersWithCountAndMax";
    }

private:

    static const int pS8_Max = 8;
    S8 pS8_CountM;
    S8 *pS8;

    static const int pU8_Max = 8;
    U8 pU8_CountM; 
    U8 *pU8;

    static const int pS16_Max = 16;
    S16 pS16_CountM;
    S16 *pS16;

    static const int pU16_Max = 16;
    U16 pU16_CountM;
    U16 *pU16;

    static const int pS32_Max = 32;
    S32 pS32_CountM;
    S32 *pS32;

    static const int pU32_Max = 32;
    U32 pU32_CountM;
    U32 *pU32;

    static const int pS64_Max = 64;
    S32 pS64_CountM;
    S64 *pS64;

    static const int pU64_Max = 64;
    U32 pU64_CountM;
    U64 *pU64;

    static const int pFl_Max = 32;
    U32 pFl_CountM;
    float *pFl;

    static const int pDl_Max = 64;
    U32 pDl_CountM;
    double *pDl;
};

class Structs : public TestClass
{
public:

    virtual void Randomize()
    {
        basicTypes.Randomize();
        basicTypeArrays.Randomize();
        basicTypeArraysWithCount.Randomize();
        basicTypePointersWithMax.Randomize();
        basicTypePointersWithCountAndMax.Randomize();
    }

    virtual const char *GetIdentity() const
    {
        return "Structs";
    }

private:

    BasicTypes basicTypes;

    BasicTypeArrays basicTypeArrays;

    BasicTypeArraysWithCount basicTypeArraysWithCount;

    BasicTypePointersWithMax basicTypePointersWithMax;

    BasicTypePointersWithCountAndMax basicTypePointersWithCountAndMax;
};

class StructArrays : public TestClass
{
public:

    virtual void Randomize()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypesArray); i++) {
            basicTypesArray[i].Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysArray); i++) {
            basicTypeArraysArray[i].Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCountArray); 
             i++) {
            basicTypeArraysWithCountArray[i].Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMaxArray); 
             i++) {
            basicTypePointersWithMaxArray[i].Randomize();
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMaxArray); i++) {
            basicTypePointersWithCountAndMaxArray[i].Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructArrays";
    }

private:
    
    BasicTypes basicTypesArray[1];

    BasicTypeArrays basicTypeArraysArray[2];

    BasicTypeArraysWithCount basicTypeArraysWithCountArray[4];

    BasicTypePointersWithMax basicTypePointersWithMaxArray[2];

    BasicTypePointersWithCountAndMax basicTypePointersWithCountAndMaxArray[1];
};

class StructPointers : public TestClass
{
public:

    inline StructPointers()
        : pBasicTypes(NULL), pBasicTypeArrays(NULL), 
          pBasicTypeArraysWithCount(NULL), pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax(NULL)
    {
    }

    virtual ~StructPointers()
    {
        delete pBasicTypes;
        delete pBasicTypeArrays;
        delete pBasicTypeArraysWithCount;
        delete pBasicTypePointersWithMax;
        delete pBasicTypePointersWithCountAndMax;
    }
    
    virtual void Randomize()
    {
        if (Util::RandomNumber() % 2) {
            pBasicTypes = new BasicTypes();
            pBasicTypes->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArrays = new BasicTypeArrays();
            pBasicTypeArrays->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArraysWithCount = new BasicTypeArraysWithCount();
            pBasicTypeArraysWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithMax = new BasicTypePointersWithMax();
            pBasicTypePointersWithMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithCountAndMax =
                new BasicTypePointersWithCountAndMax();
            pBasicTypePointersWithCountAndMax->Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructPointers";
    }

private:
    
    BasicTypes *pBasicTypes;

    BasicTypeArrays *pBasicTypeArrays;

    BasicTypeArraysWithCount *pBasicTypeArraysWithCount;

    BasicTypePointersWithMax *pBasicTypePointersWithMax;

    BasicTypePointersWithCountAndMax *pBasicTypePointersWithCountAndMax;
};

class StructArraysWithCount : public TestClass
{
public:

    inline StructArraysWithCount()
        : basicTypesArray_CountM(0),
          basicTypeArraysArray_CountM(0),
          basicTypeArraysWithCountArray_CountM(0),
          basicTypePointersWithMaxArray_CountM(0),
          basicTypePointersWithCountAndMaxArray_CountM(0)
    {
    }

    virtual void Randomize()
    {
        Rand<S8>(basicTypesArray_CountM, Util__CountOf(basicTypesArray));
        for (S8 i = 0; i < basicTypesArray_CountM; i++) {
            basicTypesArray[i].Randomize();
        }

        Rand<U8>(basicTypeArraysArray_CountM, 
                 Util__CountOf(basicTypeArraysArray));
        for (U8 i = 0; i < basicTypeArraysArray_CountM; i++) {
            basicTypeArraysArray[i].Randomize();
        }

        Rand<U16>(basicTypeArraysWithCountArray_CountM,
                  Util__CountOf(basicTypeArraysWithCountArray));
        for (U16 i = 0; i < basicTypeArraysWithCountArray_CountM; i++) {
            basicTypeArraysWithCountArray[i].Randomize();
        }

        Rand<U32>(basicTypePointersWithMaxArray_CountM,
                  Util__CountOf(basicTypePointersWithMaxArray));
        for (U32 i = 0; i < basicTypePointersWithMaxArray_CountM; i++) {
            basicTypePointersWithMaxArray[i].Randomize();
        }

        Rand<U32>(basicTypePointersWithCountAndMaxArray_CountM,
                  Util__CountOf(basicTypePointersWithCountAndMaxArray));
        for (U32 i = 0;
             i < basicTypePointersWithCountAndMaxArray_CountM; i++) {
            basicTypePointersWithCountAndMaxArray[i].Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructArraysWithCount";
    }

private:

    S8 basicTypesArray_CountM;
    BasicTypes basicTypesArray[4];

    U8 basicTypeArraysArray_CountM;
    BasicTypeArrays basicTypeArraysArray[6];

    U16 basicTypeArraysWithCountArray_CountM;
    BasicTypeArraysWithCount basicTypeArraysWithCountArray[10];

    U32 basicTypePointersWithMaxArray_CountM;
    BasicTypePointersWithMax basicTypePointersWithMaxArray[6];

    U32 basicTypePointersWithCountAndMaxArray_CountM;
    BasicTypePointersWithCountAndMax basicTypePointersWithCountAndMaxArray[4];
};

class StructPointersWithCountAndMax : public TestClass
{
public:

    inline StructPointersWithCountAndMax()
        : pBasicTypesArray_CountM(0),
          pBasicTypesArray(NULL),
          pBasicTypeArraysArray_CountM(0),
          pBasicTypeArraysArray(NULL),
          pBasicTypeArraysWithCountArray_CountM(0),
          pBasicTypeArraysWithCountArray(NULL),
          pBasicTypePointersWithMaxArray_CountM(0),
          pBasicTypePointersWithMaxArray(NULL),
          pBasicTypePointersWithCountAndMaxArray_CountM(0),
          pBasicTypePointersWithCountAndMaxArray(NULL)
    {
    }

    virtual ~StructPointersWithCountAndMax()
    {
        delete [] pBasicTypesArray;
        delete [] pBasicTypeArraysArray;
        delete [] pBasicTypeArraysWithCountArray;
        delete [] pBasicTypePointersWithMaxArray;
        delete [] pBasicTypePointersWithCountAndMaxArray;
    }

    virtual void Randomize()
    {
        Rand<S8>(pBasicTypesArray_CountM, pBasicTypesArray_Max);
        pBasicTypesArray = new BasicTypes[pBasicTypesArray_CountM];
        for (S8 i = 0; i < pBasicTypesArray_CountM; i++) {
            pBasicTypesArray[i].Randomize();
        }

        Rand<U8>(pBasicTypeArraysArray_CountM, pBasicTypeArraysArray_Max);
        pBasicTypeArraysArray = 
            new BasicTypeArrays[pBasicTypeArraysArray_CountM];
        for (U8 i = 0; i < pBasicTypeArraysArray_CountM; i++) {
            pBasicTypeArraysArray[i].Randomize();
        }

        Rand<U16>(pBasicTypeArraysWithCountArray_CountM, 
                  pBasicTypeArraysWithCountArray_Max);
        pBasicTypeArraysWithCountArray = new BasicTypeArraysWithCount
            [pBasicTypeArraysWithCountArray_CountM];
        for (U16 i = 0; i < pBasicTypeArraysWithCountArray_CountM; i++) {
            pBasicTypeArraysWithCountArray[i].Randomize();
        }

        Rand<U32>(pBasicTypePointersWithMaxArray_CountM, 
                  pBasicTypePointersWithMaxArray_Max);
        pBasicTypePointersWithMaxArray = new BasicTypePointersWithMax
            [pBasicTypePointersWithMaxArray_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithMaxArray_CountM; i++) {
            pBasicTypePointersWithMaxArray[i].Randomize();
        }

        Rand<U32>(pBasicTypePointersWithCountAndMaxArray_CountM, 
                  pBasicTypePointersWithCountAndMaxArray_Max);
        pBasicTypePointersWithCountAndMaxArray = new 
            BasicTypePointersWithCountAndMax
            [pBasicTypePointersWithCountAndMaxArray_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithCountAndMaxArray_CountM;
             i++) {
            pBasicTypePointersWithCountAndMaxArray[i].Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructPointersWithCountAndMax";
    }

private:

    static const int pBasicTypesArray_Max = 4;
    S8 pBasicTypesArray_CountM;
    BasicTypes *pBasicTypesArray;

    static const int pBasicTypeArraysArray_Max = 6;
    U8 pBasicTypeArraysArray_CountM;
    BasicTypeArrays *pBasicTypeArraysArray;

    static const int pBasicTypeArraysWithCountArray_Max = 10;
    U16 pBasicTypeArraysWithCountArray_CountM;
    BasicTypeArraysWithCount *pBasicTypeArraysWithCountArray;

    static const int pBasicTypePointersWithMaxArray_Max = 6;
    U32 pBasicTypePointersWithMaxArray_CountM;
    BasicTypePointersWithMax *pBasicTypePointersWithMaxArray;

    static const int pBasicTypePointersWithCountAndMaxArray_Max = 4;
    U32 pBasicTypePointersWithCountAndMaxArray_CountM;
    BasicTypePointersWithCountAndMax *pBasicTypePointersWithCountAndMaxArray;
};

class StructArrayOfPointers : public TestClass
{
public:

    inline StructArrayOfPointers()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypes); i++) {
            basicTypes[i] = NULL;
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArrays); i++) {
            basicTypeArrays[i] = NULL;
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCount); 
             i++) {
            basicTypeArraysWithCount[i] = NULL;
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMax); 
             i++) {
            basicTypePointersWithMax[i] = NULL;
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMax); i++) {
            basicTypePointersWithCountAndMax[i] = NULL;
        }
    }

    virtual ~StructArrayOfPointers()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypes); i++) {
            delete basicTypes[i];
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArrays); i++) {
            delete basicTypeArrays[i];
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCount); 
             i++) {
            delete basicTypeArraysWithCount[i];
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMax); 
             i++) {
            delete basicTypePointersWithMax[i];
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMax); i++) {
            delete basicTypePointersWithCountAndMax[i];
        }
    }

    virtual void Randomize()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypes); i++) {
            basicTypes[i] = new BasicTypes;
            basicTypes[i]->Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArrays); i++) {
            basicTypeArrays[i] = new BasicTypeArrays;
            basicTypeArrays[i]->Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCount); 
             i++) {
            basicTypeArraysWithCount[i] = new BasicTypeArraysWithCount;
            basicTypeArraysWithCount[i]->Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMax); 
             i++) {
            basicTypePointersWithMax[i] = new BasicTypePointersWithMax;
            basicTypePointersWithMax[i]->Randomize();
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMax); i++) {
            basicTypePointersWithCountAndMax[i] =
                new BasicTypePointersWithCountAndMax;
            basicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructArrayOfPointers";
    }

private:

    BasicTypes *basicTypes[4];

    BasicTypeArrays *basicTypeArrays[6];

    BasicTypeArraysWithCount *basicTypeArraysWithCount[10];

    BasicTypePointersWithMax *basicTypePointersWithMax[6];

    BasicTypePointersWithCountAndMax *
        basicTypePointersWithCountAndMax[4];
};

class StructArrayOfPointersWithCount : public TestClass
{
public:

    inline StructArrayOfPointersWithCount()
        : basicTypes_CountM(0),
          basicTypeArrays_CountM(0),
          basicTypeArraysWithCount_CountM(0),
          basicTypePointersWithMax_CountM(0),
          basicTypePointersWithCountAndMax_CountM(0)
    {
    }

    virtual ~StructArrayOfPointersWithCount()
    {
        for (S8 i = 0; i < basicTypes_CountM; i++) {
            delete basicTypes[i];
        }
        
        for (U8 i = 0; i < basicTypeArrays_CountM; i++) {
            delete basicTypeArrays[i];
        }

        for (U16 i = 0; i < basicTypeArraysWithCount_CountM; i++) {
            delete basicTypeArraysWithCount[i];
        }

        for (U32 i = 0; i < basicTypePointersWithMax_CountM; i++) {
            delete basicTypePointersWithMax[i];
        }

        for (U32 i = 0; i < basicTypePointersWithCountAndMax_CountM;
             i++) {
            delete basicTypePointersWithCountAndMax[i];
        }
    }

    virtual void Randomize()
    {
        Rand<S8>(basicTypes_CountM, Util__CountOf(basicTypes));
        for (S8 i = 0; i < basicTypes_CountM; i++) {
            basicTypes[i] = new BasicTypes;
            basicTypes[i]->Randomize();
        }
        
        Rand<U8>(basicTypeArrays_CountM,
                 Util__CountOf(basicTypeArrays));
        for (U8 i = 0; i < basicTypeArrays_CountM; i++) {
            basicTypeArrays[i] = new BasicTypeArrays;
            basicTypeArrays[i]->Randomize();
        }

        Rand<U16>(basicTypeArraysWithCount_CountM,
                  Util__CountOf(basicTypeArraysWithCount));
        for (U16 i = 0; i < basicTypeArraysWithCount_CountM; i++) {
            basicTypeArraysWithCount[i] = new BasicTypeArraysWithCount;
            basicTypeArraysWithCount[i]->Randomize();
        }

        Rand<U32>(basicTypePointersWithMax_CountM,
                  Util__CountOf(basicTypePointersWithMax));
        for (U32 i = 0; i < basicTypePointersWithMax_CountM; i++) {
            basicTypePointersWithMax[i] = new BasicTypePointersWithMax;
            basicTypePointersWithMax[i]->Randomize();
        }

        Rand<U32>(basicTypePointersWithCountAndMax_CountM,
                  Util__CountOf(basicTypePointersWithCountAndMax));
        for (U32 i = 0; i < basicTypePointersWithCountAndMax_CountM;
             i++) {
            basicTypePointersWithCountAndMax[i] =
                new BasicTypePointersWithCountAndMax;
            basicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructArrayOfPointersWithCount";
    }

private:
    
    S8 basicTypes_CountM;
    BasicTypes *basicTypes[4];

    U8 basicTypeArrays_CountM;
    BasicTypeArrays *basicTypeArrays[6];

    U16 basicTypeArraysWithCount_CountM;
    BasicTypeArraysWithCount *basicTypeArraysWithCount[10];

    U32 basicTypePointersWithMax_CountM;
    BasicTypePointersWithMax *basicTypePointersWithMax[6];

    U32 basicTypePointersWithCountAndMax_CountM;
    BasicTypePointersWithCountAndMax *
        basicTypePointersWithCountAndMax[4];
};

class StructDoublePointersWithMax : public TestClass
{
public:

    inline StructDoublePointersWithMax()
        : pBasicTypes(NULL),
          pBasicTypeArrays(NULL),
          pBasicTypeArraysWithCount(NULL),
          pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax(NULL)
    {
    }

    virtual ~StructDoublePointersWithMax()
    {
        if (pBasicTypes) {
            for (int i = 0; i < pBasicTypes_Max; i++) {
                delete pBasicTypes[i];
            }
            delete [] pBasicTypes;
        }

        if (pBasicTypeArrays) {
            for (int i = 0; i < pBasicTypeArrays_Max; i++) {
                delete pBasicTypeArrays[i];
            }
            delete [] pBasicTypeArrays;
        }

        if (pBasicTypeArraysWithCount) {
            for (int i = 0; i < pBasicTypeArraysWithCount_Max; i++) {
                delete pBasicTypeArraysWithCount[i];
            }
            delete [] pBasicTypeArraysWithCount;
        }

        if (pBasicTypePointersWithMax) {
            for (int i = 0; i < pBasicTypePointersWithMax_Max; i++) {
                delete pBasicTypePointersWithMax[i];
            }
            delete [] pBasicTypePointersWithMax;
        }

        if (pBasicTypePointersWithCountAndMax) {
            for (int i = 0; i < pBasicTypePointersWithCountAndMax_Max;
                 i++) {
                delete pBasicTypePointersWithCountAndMax[i];
            }
            delete [] pBasicTypePointersWithCountAndMax;
        }
    }

    virtual void Randomize()
    {
        pBasicTypes = new BasicTypes * [pBasicTypes_Max];
        for (int i = 0; i < pBasicTypes_Max; i++) {
            pBasicTypes[i] = new BasicTypes;
            pBasicTypes[i]->Randomize();
        }

        pBasicTypeArrays = new BasicTypeArrays *
            [pBasicTypeArrays_Max];
        for (int i = 0; i < pBasicTypeArrays_Max; i++) {
            pBasicTypeArrays[i] = new BasicTypeArrays;
            pBasicTypeArrays[i]->Randomize();
        }

        pBasicTypeArraysWithCount = new BasicTypeArraysWithCount *
            [pBasicTypeArraysWithCount_Max];
        for (int i = 0; i < pBasicTypeArraysWithCount_Max; i++) {
            pBasicTypeArraysWithCount[i] = new
                BasicTypeArraysWithCount;
            pBasicTypeArraysWithCount[i]->Randomize();
        }

        pBasicTypePointersWithMax = 
            new BasicTypePointersWithMax *
            [pBasicTypePointersWithMax_Max];
        for (int i = 0; i < pBasicTypePointersWithMax_Max; i++) {
            pBasicTypePointersWithMax[i] = new
                BasicTypePointersWithMax;
            pBasicTypePointersWithMax[i]->Randomize();
        }

        pBasicTypePointersWithCountAndMax = 
            new BasicTypePointersWithCountAndMax *
            [pBasicTypePointersWithCountAndMax_Max];
        for (int i = 0; i < pBasicTypePointersWithCountAndMax_Max;
             i++) {
            pBasicTypePointersWithCountAndMax[i] = new
                BasicTypePointersWithCountAndMax;
            pBasicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructDoublePointersWithMax";
    }

private:
    
    static const int pBasicTypes_Max = 4;
    BasicTypes **pBasicTypes;

    static const int pBasicTypeArrays_Max = 6;
    BasicTypeArrays **pBasicTypeArrays;

    static const int pBasicTypeArraysWithCount_Max = 10;
    BasicTypeArraysWithCount **pBasicTypeArraysWithCount;

    static const int pBasicTypePointersWithMax_Max = 6;
    BasicTypePointersWithMax **pBasicTypePointersWithMax;

    static const int pBasicTypePointersWithCountAndMax_Max = 4;
    BasicTypePointersWithCountAndMax **pBasicTypePointersWithCountAndMax;
};

class StructDoublePointersWithCountAndMax : public TestClass
{
public:

    inline StructDoublePointersWithCountAndMax()
        : pBasicTypes_CountM(0),
          pBasicTypes(NULL),
          pBasicTypeArrays_CountM(0),
          pBasicTypeArrays(NULL),
          pBasicTypeArraysWithCount_CountM(0),
          pBasicTypeArraysWithCount(NULL),
          pBasicTypePointersWithMax_CountM(0),
          pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax_CountM(0),
          pBasicTypePointersWithCountAndMax(NULL)
    {
    }

    virtual ~StructDoublePointersWithCountAndMax()
    {
        for (S8 i = 0; i < pBasicTypes_CountM; i++) {
            delete pBasicTypes[i];
        }
        delete [] pBasicTypes;

        for (U8 i = 0; i < pBasicTypeArrays_CountM; i++) {
            delete pBasicTypeArrays[i];
        }
        delete [] pBasicTypeArrays;

        for (U16 i = 0; i < pBasicTypeArraysWithCount_CountM; i++) {
            delete pBasicTypeArraysWithCount[i];
        }
        delete [] pBasicTypeArraysWithCount;

        for (U32 i = 0; i < pBasicTypePointersWithMax_CountM; i++) {
            delete pBasicTypePointersWithMax[i];
        }
        delete [] pBasicTypePointersWithMax;

        for (U32 i = 0; i < pBasicTypePointersWithCountAndMax_CountM;
             i++) {
            delete pBasicTypePointersWithCountAndMax[i];
        }
        delete [] pBasicTypePointersWithCountAndMax;
    }

    virtual void Randomize()
    {
        Rand<S8>(pBasicTypes_CountM, pBasicTypes_Max);
        pBasicTypes = new BasicTypes * [pBasicTypes_CountM];
        for (S8 i = 0; i < pBasicTypes_CountM; i++) {
            pBasicTypes[i] = new BasicTypes;
            pBasicTypes[i]->Randomize();
        }

        Rand<U8>(pBasicTypeArrays_CountM, pBasicTypeArrays_Max);
        pBasicTypeArrays = new BasicTypeArrays * 
            [pBasicTypeArrays_CountM];
        for (U8 i = 0; i < pBasicTypeArrays_CountM; i++) {
            pBasicTypeArrays[i] = new BasicTypeArrays;
            pBasicTypeArrays[i]->Randomize();
        }

        Rand<U16>(pBasicTypeArraysWithCount_CountM,
                  pBasicTypeArraysWithCount_Max);
        pBasicTypeArraysWithCount = new BasicTypeArraysWithCount * 
            [pBasicTypeArraysWithCount_CountM];
        for (U16 i = 0; i < pBasicTypeArraysWithCount_CountM; i++) {
            pBasicTypeArraysWithCount[i] = 
                new BasicTypeArraysWithCount;
            pBasicTypeArraysWithCount[i]->Randomize();
        }

        Rand<U32>(pBasicTypePointersWithMax_CountM,
                  pBasicTypePointersWithMax_Max);
        pBasicTypePointersWithMax = new BasicTypePointersWithMax * 
            [pBasicTypePointersWithMax_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithMax_CountM; i++) {
            pBasicTypePointersWithMax[i] =
                new BasicTypePointersWithMax;
            pBasicTypePointersWithMax[i]->Randomize();
        }

        Rand<U32>(pBasicTypePointersWithCountAndMax_CountM,
                  pBasicTypePointersWithCountAndMax_Max);
        pBasicTypePointersWithCountAndMax = 
            new BasicTypePointersWithCountAndMax * 
            [pBasicTypePointersWithCountAndMax_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithCountAndMax_CountM;
             i++) {
            pBasicTypePointersWithCountAndMax[i] =
                new BasicTypePointersWithCountAndMax;
            pBasicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "StructDoublePointersWithCountAndMax";
    }

private:
    
    static const int pBasicTypes_Max = 4;
    S8 pBasicTypes_CountM;
    BasicTypes **pBasicTypes;

    static const int pBasicTypeArrays_Max = 6;
    U8 pBasicTypeArrays_CountM;
    BasicTypeArrays **pBasicTypeArrays;

    static const int pBasicTypeArraysWithCount_Max = 10;
    U16 pBasicTypeArraysWithCount_CountM;
    BasicTypeArraysWithCount **pBasicTypeArraysWithCount;

    static const int pBasicTypePointersWithMax_Max = 6;
    U32 pBasicTypePointersWithMax_CountM;
    BasicTypePointersWithMax **pBasicTypePointersWithMax;

    static const int pBasicTypePointersWithCountAndMax_Max = 4;
    U32 pBasicTypePointersWithCountAndMax_CountM;
    BasicTypePointersWithCountAndMax **pBasicTypePointersWithCountAndMax;
};

class DeepInnerBasicType : public TestClass
{
public:

    virtual void Randomize()
    {
        Rand<U64>(outer.middle.inner.u64);
    }

    virtual const char *GetIdentity() const
    {
        return "DeepInnerBasicType";
    }

private:

    struct {
        class {
        public:
            struct {
                U64 u64;
            } inner;
        } middle;
    } outer;
};

class DeepInnerStructAndHierarchy : public TestClass
{
public:

    virtual void Randomize()
    {
        outer.middle.Randomize();
    }

    virtual const char *GetIdentity() const
    {
        return "DeepInnerStructAndHierarchy";
    }

private:

    struct {
        class {
        public:

            void Randomize()
            {
                inner.upper.Randomize();
                inner.middle.Randomize();
                inner.bottom.Randomize();
            }

            struct {

                class Upper
                {
                public:

                    void Randomize()
                    {
                        upperOuter.Randomize();
                    }

                    class UpperOuter
                    {
                    public:

                        void Randomize()
                        {
                            serialized.middle.Randomize();

                            serialized.bottom.Randomize();

                            RandReal<double>(transient_double);
                        }

                        class Middle
                        {
                        public:

                            void Randomize()
                            {
                                Rand<U8>(u8);
                            }
                            
                        private:

                            U8 u8;
                        };

                        class Bottom : public Middle
                        {
                        public:

                            void Randomize()
                            {
                                this->Middle::Randomize();

                                RandReal<float>(serialized.fl);
                            }

                        private:

                            struct {
                                float fl;
                            } serialized;
                        };

                        struct {
                            Middle middle;
                            Bottom bottom;
                        } serialized;

                        double transient_double;
                    };

                    UpperOuter upperOuter;
                };

                class Middle : public Upper
                {
                public:

                    void Randomize()
                    {
                        this->Upper::Randomize();

                        Rand<U16>(serialized.u16);
                        
                        RandReal<double>(transient_double);
                    }

                private:
                    
                    struct {
                        U16 u16;
                    } serialized;

                    double transient_double;
                };
                
                class Bottom : public Middle
                {
                public:

                    void Randomize()
                    {
                        this->Middle::Randomize();

                        Rand<U32>(serialized.u32);

                        RandReal<double>(transient_double);
                    }

                private:
                    
                    struct {
                        U32 u32;
                    } serialized;

                    double transient_double;
                };
                
                Upper upper;
                Middle middle;
                Bottom bottom;
            } inner;
        } middle;
    } outer;
};

class BigArray : public TestClass
{
public:

    virtual void Randomize()
    {
        U32 array_count;
        Rand<U32>(array_count, Util__CountOf(array) - 1);

        for (U32 i = 0; i < array_count; i++) {
            array[i] = (Util::RandomNumber() % ('z' - 'a')) + 'a';
        }
        array[array_count] = 0;

        for (U32 i = 0; i < Util__CountOf(array2); i++) {
            RandReal<float>(array2[i]);
        }
    }

    virtual const char *GetIdentity() const
    {
        return "BigArray";
    }

private:

    U8 array[1005];

    float array2[1020];
};

class SimpleStructArray : public TestClass
{
public:
    
    virtual void Randomize()
    {
        Rand<U32>(simpleStructs_CountM, Util__CountOf(simpleStructs));

        for (U32 i = 0; i < simpleStructs_CountM; i++) {
            Rand<U32>(simpleStructs[i].val1);
            Rand<U32>(simpleStructs[i].val2);
            Rand<U32>(simpleStructs[i].val3);
        }
    }

    virtual const char *GetIdentity() const
    {
        return "SimpleStructArray";
    }

private:

    class SimpleStruct
    {
    public:

        U32 val1, val2, val3;
    };

    U32 simpleStructs_CountM;
    SimpleStruct simpleStructs[100];
};

class Optional : public TestClass
{
public:
    
    Optional()
        : pBasicTypes(NULL), pBasicTypeArrays(NULL), 
          pBasicTypeArraysWithCount(NULL), pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax(NULL), pStructs(NULL),
          pStructArrays(NULL), pStructPointers(NULL), 
          pStructArraysWithCount(NULL), pStructPointersWithCountAndMax(NULL), 
          pStructArrayOfPointers(NULL), pStructArrayOfPointersWithCount(NULL), 
          pStructDoublePointersWithMax(NULL), 
          pStructDoublePointersWithCountAndMax(NULL),
          pDeepInnerBasicType(NULL), pDeepInnerStructAndHierarchy(NULL), 
          pBigArray(NULL), pSimpleStructArray(NULL)
    {
    }

    virtual ~Optional()
    {
        delete pBasicTypes;
        delete pBasicTypeArrays;
        delete pBasicTypeArraysWithCount;
        delete pBasicTypePointersWithMax;
        delete pBasicTypePointersWithCountAndMax;
        delete pStructs;
        delete pStructArrays;
        delete pStructPointers;
        delete pStructArraysWithCount;
        delete pStructPointersWithCountAndMax;
        delete pStructArrayOfPointers;
        delete pStructArrayOfPointersWithCount;
        delete pStructDoublePointersWithMax;
        delete pStructDoublePointersWithCountAndMax;
        delete pDeepInnerBasicType;
        delete pDeepInnerStructAndHierarchy;
        delete pBigArray;
        delete pSimpleStructArray;
    }

    virtual void Randomize()
    {
        if (Util::RandomNumber() % 2) {
            pBasicTypes = new BasicTypes();
            pBasicTypes->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArrays = new BasicTypeArrays();
            pBasicTypeArrays->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArraysWithCount = new BasicTypeArraysWithCount();
            pBasicTypeArraysWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithMax = new BasicTypePointersWithMax();
            pBasicTypePointersWithMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithCountAndMax = new
                BasicTypePointersWithCountAndMax();
            pBasicTypePointersWithCountAndMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructs = new Structs();
            pStructs->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArrays = new StructArrays();
            pStructArrays->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructPointers = new StructPointers();
            pStructPointers->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArraysWithCount = new StructArraysWithCount();
            pStructArraysWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructPointersWithCountAndMax = new
                StructPointersWithCountAndMax();
            pStructPointersWithCountAndMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArrayOfPointers = new StructArrayOfPointers();
            pStructArrayOfPointers->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArrayOfPointersWithCount = new
                StructArrayOfPointersWithCount();
            pStructArrayOfPointersWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructDoublePointersWithMax = new StructDoublePointersWithMax();
            pStructDoublePointersWithMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructDoublePointersWithCountAndMax = new
                StructDoublePointersWithCountAndMax();
            pStructDoublePointersWithCountAndMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pDeepInnerBasicType = new DeepInnerBasicType();
            pDeepInnerBasicType->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pDeepInnerStructAndHierarchy = new DeepInnerStructAndHierarchy();
            pDeepInnerStructAndHierarchy->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBigArray = new BigArray();
            pBigArray->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pSimpleStructArray = new SimpleStructArray();
            pSimpleStructArray->Randomize();
        }
    }

    virtual const char *GetIdentity() const
    {
        return "Optional";
    }

private:

    BasicTypes *pBasicTypes;
    BasicTypeArrays *pBasicTypeArrays;
    BasicTypeArraysWithCount *pBasicTypeArraysWithCount;
    BasicTypePointersWithMax *pBasicTypePointersWithMax;
    BasicTypePointersWithCountAndMax *pBasicTypePointersWithCountAndMax;
    Structs *pStructs;
    StructArrays *pStructArrays;
    StructPointers *pStructPointers;
    StructArraysWithCount *pStructArraysWithCount;
    StructPointersWithCountAndMax *pStructPointersWithCountAndMax;
    StructArrayOfPointers *pStructArrayOfPointers;
    StructArrayOfPointersWithCount *pStructArrayOfPointersWithCount;
    StructDoublePointersWithMax *pStructDoublePointersWithMax;
    StructDoublePointersWithCountAndMax *pStructDoublePointersWithCountAndMax;
    DeepInnerBasicType *pDeepInnerBasicType;
    DeepInnerStructAndHierarchy *pDeepInnerStructAndHierarchy;
    BigArray *pBigArray;
    SimpleStructArray *pSimpleStructArray;
};

class LocalTestClass
{
public:

    virtual ~LocalTestClass() { }

    virtual void Randomize() { }
};

class LocalBasicTypes : public LocalTestClass
{
public:

    // The actual value of fRandomize does not matter
    virtual void Randomize()
    {
        Rand<S8>(s8);
        Rand<U8>(u8);
        Rand<S16>(s16);
        Rand<U16>(u16);
        Rand<S32>(s32);
        Rand<U32>(u32);
        Rand<S64>(s64);
        Rand<U64>(u64);
        RandReal<float>(fl);
        RandReal<double>(dl);
    }

private:
    
    S8 s8;

    U8 u8;

    S16 s16;

    U16 u16;

    S32 s32;

    U32 u32;

    S64 s64;

    U64 u64;

    float fl;

    double dl;
};

class LocalBasicTypeArrays : public LocalTestClass
{
public:
    
    // The actual value of fRandomize does not matter
    virtual void Randomize()
    {
        RandArray<S8>(s8, Util__CountOf(s8));
        RandArray<U8>(u8, Util__CountOf(u8));
        RandArray<S16>(s16, Util__CountOf(s16));
        RandArray<U16>(u16, Util__CountOf(u16));
        RandArray<S32>(s32, Util__CountOf(s32));
        RandArray<U32>(u32, Util__CountOf(u32));
        RandArray<S64>(s64, Util__CountOf(s64));
        RandArray<U64>(u64, Util__CountOf(u64));
        RandRealArray<float>(fl, Util__CountOf(fl));
        RandRealArray<double>(dl, Util__CountOf(dl));
    }

private:

    S8 s8[8];

    U8 u8[8];

    S16 s16[16];

    U16 u16[16];

    S32 s32[32];

    U32 u32[32];

    S64 s64[64];

    U64 u64[64];

    float fl[10];

    double dl[10];
};

class LocalBasicTypeArraysWithCount : public LocalTestClass
{
public:

    inline LocalBasicTypeArraysWithCount()
        : s8_CountM(0), u8_CountM(0), s16_CountM(0), u16_CountM(0),
          s32_CountM(0), u32_CountM(0), s64_CountM(0), u64_CountM(0),
          fl_CountM(0), dl_CountM(0)
    {
    }

    virtual void Randomize()
    {
        RandArray<S8, S8>(s8, s8_CountM, Util__CountOf(s8));
        RandArray<U8, U8>(u8, u8_CountM, Util__CountOf(u8));
        RandArray<S16, S16>(s16, s16_CountM, Util__CountOf(s16));
        RandArray<U16, U16>(u16, u16_CountM, Util__CountOf(u16));
        RandArray<S32, S32>(s32, s32_CountM, Util__CountOf(s32));
        RandArray<U32, U32>(u32, u32_CountM, Util__CountOf(u32));
        RandArray<S64, S32>(s64, s64_CountM, Util__CountOf(s64));
        RandArray<U64, U32>(u64, u64_CountM, Util__CountOf(u64));
        RandRealArray<float, U32>(fl, fl_CountM, Util__CountOf(fl));
        RandRealArray<double, U32>(dl, dl_CountM, Util__CountOf(dl));
    }

private:

    S8 s8_CountM;
    S8 s8[8];

    U8 u8_CountM; 
    U8 u8[8];

    S16 s16_CountM;
    S16 s16[16];

    U16 u16_CountM;
    U16 u16[16];

    S32 s32_CountM;
    S32 s32[32];

    U32 u32_CountM;
    U32 u32[32];

    S32 s64_CountM;
    S64 s64[64];

    U32 u64_CountM;
    U64 u64[64];

    U32 fl_CountM;
    float fl[10];

    U32 dl_CountM;
    double dl[10];
};

class LocalBasicTypePointersWithMax : public LocalTestClass
{
public:

    inline LocalBasicTypePointersWithMax()
        : pS8(NULL), pU8(NULL), pS16(NULL), pU16(NULL), pS32(NULL), pU32(NULL),
          pS64(NULL), pU64(NULL), pFl(NULL), pDl(NULL)
    {
    }

    virtual ~LocalBasicTypePointersWithMax()
    {
        delete [] pS8;
        delete [] pU8;
        delete [] pS16;
        delete [] pU16;
        delete [] pS32;
        delete [] pU32;
        delete [] pS64;
        delete [] pU64;
        delete [] pFl;
        delete [] pDl;
    }

    virtual void Randomize()
    {
        CreateRandArray<S8>(pS8, pS8_Max);
        CreateRandArray<U8>(pU8, pU8_Max);
        CreateRandArray<S16>(pS16, pS16_Max);
        CreateRandArray<U16>(pU16, pU16_Max);
        CreateRandArray<S32>(pS32, pS32_Max);
        CreateRandArray<U32>(pU32, pU32_Max);
        CreateRandArray<S64>(pS64, pS64_Max);
        CreateRandArray<U64>(pU64, pU64_Max);
        CreateRandRealArray<float>(pFl, pFl_Max);
        CreateRandRealArray<double>(pDl, pDl_Max);
    }

private:

    static const int pS8_Max = 8;
    S8 *pS8;

    static const int pU8_Max = 8;
    U8 *pU8;

    static const int pS16_Max = 16;
    S16 *pS16;

    static const int pU16_Max = 16;
    U16 *pU16;

    static const int pS32_Max = 32;
    S32 *pS32;

    static const int pU32_Max = 32;
    U32 *pU32;

    static const int pS64_Max = 64;
    S64 *pS64;

    static const int pU64_Max = 64;
    U64 *pU64;

    static const int pFl_Max = 32;
    float *pFl;

    static const int pDl_Max = 64;
    double *pDl;
};

class LocalBasicTypePointersWithCountAndMax : public LocalTestClass
{
public:

    inline LocalBasicTypePointersWithCountAndMax()
        : pS8_CountM(0), pS8(NULL), pU8_CountM(0), pU8(NULL), 
          pS16_CountM(0), pS16(NULL), pU16_CountM(0), pU16(NULL), 
          pS32_CountM(0), pS32(NULL), pU32_CountM(0), pU32(NULL), 
          pS64_CountM(0), pS64(NULL), pU64_CountM(0), pU64(NULL), 
          pFl_CountM(0), pFl(NULL), pDl_CountM(0), pDl(NULL)
    {
    }

    virtual ~LocalBasicTypePointersWithCountAndMax()
    {
        delete [] pS8;
        delete [] pU8;
        delete [] pS16;
        delete [] pU16;
        delete [] pS32;
        delete [] pU32;
        delete [] pS64;
        delete [] pU64;
        delete [] pFl;
        delete [] pDl;
    }

    virtual void Randomize()
    {
        CreateRandArray<S8, S8>(pS8, pS8_CountM, pS8_Max);
        CreateRandArray<U8, U8>(pU8, pU8_CountM, pU8_Max);
        CreateRandArray<S16, S16>(pS16, pS16_CountM, pS16_Max);
        CreateRandArray<U16, U16>(pU16, pU16_CountM, pU16_Max);
        CreateRandArray<S32, S32>(pS32, pS32_CountM, pS32_Max);
        CreateRandArray<U32, U32>(pU32, pU32_CountM, pU32_Max);
        CreateRandArray<S64, S32>(pS64, pS64_CountM, pS64_Max);
        CreateRandArray<U64, U32>(pU64, pU64_CountM, pU64_Max);
        CreateRandRealArray<float, U32>(pFl, pFl_CountM, pFl_Max);
        CreateRandRealArray<double, U32>(pDl, pDl_CountM, pDl_Max);
    }

private:

    static const int pS8_Max = 8;
    S8 pS8_CountM;
    S8 *pS8;

    static const int pU8_Max = 8;
    U8 pU8_CountM; 
    U8 *pU8;

    static const int pS16_Max = 16;
    S16 pS16_CountM;
    S16 *pS16;

    static const int pU16_Max = 16;
    U16 pU16_CountM;
    U16 *pU16;

    static const int pS32_Max = 32;
    S32 pS32_CountM;
    S32 *pS32;

    static const int pU32_Max = 32;
    U32 pU32_CountM;
    U32 *pU32;

    static const int pS64_Max = 64;
    S32 pS64_CountM;
    S64 *pS64;

    static const int pU64_Max = 64;
    U32 pU64_CountM;
    U64 *pU64;

    static const int pFl_Max = 32;
    U32 pFl_CountM;
    float *pFl;

    static const int pDl_Max = 64;
    U32 pDl_CountM;
    double *pDl;
};

class LocalStructs : public LocalTestClass
{
public:

    virtual void Randomize()
    {
        basicTypes.Randomize();
        basicTypeArrays.Randomize();
        basicTypeArraysWithCount.Randomize();
        basicTypePointersWithMax.Randomize();
        basicTypePointersWithCountAndMax.Randomize();
    }

private:

    LocalBasicTypes basicTypes;

    BasicTypeArrays basicTypeArrays;

    LocalBasicTypeArraysWithCount basicTypeArraysWithCount;

    BasicTypePointersWithMax basicTypePointersWithMax;

    LocalBasicTypePointersWithCountAndMax basicTypePointersWithCountAndMax;
};

class LocalStructArrays : public LocalTestClass
{
public:

    virtual void Randomize()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypesArray); i++) {
            basicTypesArray[i].Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysArray); i++) {
            basicTypeArraysArray[i].Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCountArray); 
             i++) {
            basicTypeArraysWithCountArray[i].Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMaxArray); 
             i++) {
            basicTypePointersWithMaxArray[i].Randomize();
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMaxArray); i++) {
            basicTypePointersWithCountAndMaxArray[i].Randomize();
        }
    }

private:
    
    BasicTypes basicTypesArray[1];

    LocalBasicTypeArrays basicTypeArraysArray[2];

    BasicTypeArraysWithCount basicTypeArraysWithCountArray[4];

    LocalBasicTypePointersWithMax basicTypePointersWithMaxArray[2];

    BasicTypePointersWithCountAndMax basicTypePointersWithCountAndMaxArray[1];
};

class LocalStructPointers : public LocalTestClass
{
public:

    inline LocalStructPointers()
        : pBasicTypes(NULL), pBasicTypeArrays(NULL), 
          pBasicTypeArraysWithCount(NULL), pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax(NULL)
    {
    }

    virtual ~LocalStructPointers()
    {
        delete pBasicTypes;
        delete pBasicTypeArrays;
        delete pBasicTypeArraysWithCount;
        delete pBasicTypePointersWithMax;
        delete pBasicTypePointersWithCountAndMax;
    }
    
    virtual void Randomize()
    {
        if (Util::RandomNumber() % 2) {
            pBasicTypes = new LocalBasicTypes();
            pBasicTypes->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArrays = new LocalBasicTypeArrays();
            pBasicTypeArrays->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArraysWithCount = new BasicTypeArraysWithCount();
            pBasicTypeArraysWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithMax = new LocalBasicTypePointersWithMax();
            pBasicTypePointersWithMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithCountAndMax =
                new BasicTypePointersWithCountAndMax();
            pBasicTypePointersWithCountAndMax->Randomize();
        }
    }

private:
    
    LocalBasicTypes *pBasicTypes;
    
    LocalBasicTypeArrays *pBasicTypeArrays;

    BasicTypeArraysWithCount *pBasicTypeArraysWithCount;

    LocalBasicTypePointersWithMax *pBasicTypePointersWithMax;

    BasicTypePointersWithCountAndMax *pBasicTypePointersWithCountAndMax;
};

class LocalStructArraysWithCount : public LocalTestClass
{
public:

    inline LocalStructArraysWithCount()
        : basicTypesArray_CountM(0),
          basicTypeArraysArray_CountM(0),
          basicTypeArraysWithCountArray_CountM(0),
          basicTypePointersWithMaxArray_CountM(0),
          basicTypePointersWithCountAndMaxArray_CountM(0)
    {
    }

    virtual void Randomize()
    {
        Rand<S8>(basicTypesArray_CountM, Util__CountOf(basicTypesArray));
        for (S8 i = 0; i < basicTypesArray_CountM; i++) {
            basicTypesArray[i].Randomize();
        }

        Rand<U8>(basicTypeArraysArray_CountM, 
                 Util__CountOf(basicTypeArraysArray));
        for (U8 i = 0; i < basicTypeArraysArray_CountM; i++) {
            basicTypeArraysArray[i].Randomize();
        }

        Rand<U16>(basicTypeArraysWithCountArray_CountM,
                  Util__CountOf(basicTypeArraysWithCountArray));
        for (U16 i = 0; i < basicTypeArraysWithCountArray_CountM; i++) {
            basicTypeArraysWithCountArray[i].Randomize();
        }

        Rand<U32>(basicTypePointersWithMaxArray_CountM,
                  Util__CountOf(basicTypePointersWithMaxArray));
        for (U32 i = 0; i < basicTypePointersWithMaxArray_CountM; i++) {
            basicTypePointersWithMaxArray[i].Randomize();
        }

        Rand<U32>(basicTypePointersWithCountAndMaxArray_CountM,
                  Util__CountOf(basicTypePointersWithCountAndMaxArray));
        for (U32 i = 0;
             i < basicTypePointersWithCountAndMaxArray_CountM; i++) {
            basicTypePointersWithCountAndMaxArray[i].Randomize();
        }
    }

private:

    S8 basicTypesArray_CountM;
    BasicTypes basicTypesArray[4];

    U8 basicTypeArraysArray_CountM;
    BasicTypeArrays basicTypeArraysArray[6];

    U16 basicTypeArraysWithCountArray_CountM;
    LocalBasicTypeArraysWithCount basicTypeArraysWithCountArray[10];

    U32 basicTypePointersWithMaxArray_CountM;
    BasicTypePointersWithMax basicTypePointersWithMaxArray[6];

    U32 basicTypePointersWithCountAndMaxArray_CountM;
    LocalBasicTypePointersWithCountAndMax 
        basicTypePointersWithCountAndMaxArray[4];
};

class LocalStructPointersWithCountAndMax : public LocalTestClass
{
public:

    inline LocalStructPointersWithCountAndMax()
        : pBasicTypesArray_CountM(0),
          pBasicTypesArray(NULL),
          pBasicTypeArraysArray_CountM(0),
          pBasicTypeArraysArray(NULL),
          pBasicTypeArraysWithCountArray_CountM(0),
          pBasicTypeArraysWithCountArray(NULL),
          pBasicTypePointersWithMaxArray_CountM(0),
          pBasicTypePointersWithMaxArray(NULL),
          pBasicTypePointersWithCountAndMaxArray_CountM(0),
          pBasicTypePointersWithCountAndMaxArray(NULL)
    {
    }

    virtual ~LocalStructPointersWithCountAndMax()
    {
        delete [] pBasicTypesArray;
        delete [] pBasicTypeArraysArray;
        delete [] pBasicTypeArraysWithCountArray;
        delete [] pBasicTypePointersWithMaxArray;
        delete [] pBasicTypePointersWithCountAndMaxArray;
    }

    virtual void Randomize()
    {
        Rand<S8>(pBasicTypesArray_CountM, pBasicTypesArray_Max);
        pBasicTypesArray = new BasicTypes[pBasicTypesArray_CountM];
        for (S8 i = 0; i < pBasicTypesArray_CountM; i++) {
            pBasicTypesArray[i].Randomize();
        }

        Rand<U8>(pBasicTypeArraysArray_CountM, pBasicTypeArraysArray_Max);
        pBasicTypeArraysArray = 
            new BasicTypeArrays[pBasicTypeArraysArray_CountM];
        for (U8 i = 0; i < pBasicTypeArraysArray_CountM; i++) {
            pBasicTypeArraysArray[i].Randomize();
        }

        Rand<U16>(pBasicTypeArraysWithCountArray_CountM, 
                  pBasicTypeArraysWithCountArray_Max);
        pBasicTypeArraysWithCountArray = new LocalBasicTypeArraysWithCount
            [pBasicTypeArraysWithCountArray_CountM];
        for (U16 i = 0; i < pBasicTypeArraysWithCountArray_CountM; i++) {
            pBasicTypeArraysWithCountArray[i].Randomize();
        }

        Rand<U32>(pBasicTypePointersWithMaxArray_CountM, 
                  pBasicTypePointersWithMaxArray_Max);
        pBasicTypePointersWithMaxArray = new BasicTypePointersWithMax
            [pBasicTypePointersWithMaxArray_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithMaxArray_CountM; i++) {
            pBasicTypePointersWithMaxArray[i].Randomize();
        }

        Rand<U32>(pBasicTypePointersWithCountAndMaxArray_CountM, 
                  pBasicTypePointersWithCountAndMaxArray_Max);
        pBasicTypePointersWithCountAndMaxArray = new 
            BasicTypePointersWithCountAndMax
            [pBasicTypePointersWithCountAndMaxArray_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithCountAndMaxArray_CountM;
             i++) {
            pBasicTypePointersWithCountAndMaxArray[i].Randomize();
        }
    }

private:

    static const int pBasicTypesArray_Max = 4;
    S8 pBasicTypesArray_CountM;
    BasicTypes *pBasicTypesArray;

    static const int pBasicTypeArraysArray_Max = 6;
    U8 pBasicTypeArraysArray_CountM;
    BasicTypeArrays *pBasicTypeArraysArray;

    static const int pBasicTypeArraysWithCountArray_Max = 10;
    U16 pBasicTypeArraysWithCountArray_CountM;
    LocalBasicTypeArraysWithCount *pBasicTypeArraysWithCountArray;

    static const int pBasicTypePointersWithMaxArray_Max = 6;
    U32 pBasicTypePointersWithMaxArray_CountM;
    BasicTypePointersWithMax *pBasicTypePointersWithMaxArray;

    static const int pBasicTypePointersWithCountAndMaxArray_Max = 4;
    U32 pBasicTypePointersWithCountAndMaxArray_CountM;
    BasicTypePointersWithCountAndMax *pBasicTypePointersWithCountAndMaxArray;
};

class LocalStructArrayOfPointers : public LocalTestClass
{
public:

    inline LocalStructArrayOfPointers()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypes); i++) {
            basicTypes[i] = NULL;
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArrays); i++) {
            basicTypeArrays[i] = NULL;
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCount); 
             i++) {
            basicTypeArraysWithCount[i] = NULL;
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMax); 
             i++) {
            basicTypePointersWithMax[i] = NULL;
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMax); i++) {
            basicTypePointersWithCountAndMax[i] = NULL;
        }
    }

    virtual ~LocalStructArrayOfPointers()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypes); i++) {
            delete basicTypes[i];
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArrays); i++) {
            delete basicTypeArrays[i];
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCount); 
             i++) {
            delete basicTypeArraysWithCount[i];
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMax); 
             i++) {
            delete basicTypePointersWithMax[i];
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMax); i++) {
            delete basicTypePointersWithCountAndMax[i];
        }
    }

    virtual void Randomize()
    {
        for (U32 i = 0; i < Util__CountOf(basicTypes); i++) {
            basicTypes[i] = new LocalBasicTypes;
            basicTypes[i]->Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArrays); i++) {
            basicTypeArrays[i] = new LocalBasicTypeArrays;
            basicTypeArrays[i]->Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypeArraysWithCount); 
             i++) {
            basicTypeArraysWithCount[i] = new BasicTypeArraysWithCount;
            basicTypeArraysWithCount[i]->Randomize();
        }

        for (U32 i = 0; i < Util__CountOf(basicTypePointersWithMax); 
             i++) {
            basicTypePointersWithMax[i] = new LocalBasicTypePointersWithMax;
            basicTypePointersWithMax[i]->Randomize();
        }

        for (U32 i = 0; 
             i < Util__CountOf(basicTypePointersWithCountAndMax); i++) {
            basicTypePointersWithCountAndMax[i] =
                new LocalBasicTypePointersWithCountAndMax;
            basicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

private:

    LocalBasicTypes *basicTypes[4];

    LocalBasicTypeArrays *basicTypeArrays[6];

    BasicTypeArraysWithCount *basicTypeArraysWithCount[10];

    LocalBasicTypePointersWithMax *basicTypePointersWithMax[6];

    LocalBasicTypePointersWithCountAndMax *
        basicTypePointersWithCountAndMax[4];
};

class LocalStructArrayOfPointersWithCount : public LocalTestClass
{
public:

    inline LocalStructArrayOfPointersWithCount()
        : basicTypes_CountM(0),
          basicTypeArrays_CountM(0),
          basicTypeArraysWithCount_CountM(0),
          basicTypePointersWithMax_CountM(0),
          basicTypePointersWithCountAndMax_CountM(0)
    {
    }

    virtual ~LocalStructArrayOfPointersWithCount()
    {
        for (S8 i = 0; i < basicTypes_CountM; i++) {
            delete basicTypes[i];
        }
        
        for (U8 i = 0; i < basicTypeArrays_CountM; i++) {
            delete basicTypeArrays[i];
        }

        for (U16 i = 0; i < basicTypeArraysWithCount_CountM; i++) {
            delete basicTypeArraysWithCount[i];
        }

        for (U32 i = 0; i < basicTypePointersWithMax_CountM; i++) {
            delete basicTypePointersWithMax[i];
        }

        for (U32 i = 0; i < basicTypePointersWithCountAndMax_CountM;
             i++) {
            delete basicTypePointersWithCountAndMax[i];
        }
    }

    virtual void Randomize()
    {
        Rand<S8>(basicTypes_CountM, Util__CountOf(basicTypes));
        for (S8 i = 0; i < basicTypes_CountM; i++) {
            basicTypes[i] = new LocalBasicTypes;
            basicTypes[i]->Randomize();
        }
        
        Rand<U8>(basicTypeArrays_CountM,
                 Util__CountOf(basicTypeArrays));
        for (U8 i = 0; i < basicTypeArrays_CountM; i++) {
            basicTypeArrays[i] = new BasicTypeArrays;
            basicTypeArrays[i]->Randomize();
        }

        Rand<U16>(basicTypeArraysWithCount_CountM,
                  Util__CountOf(basicTypeArraysWithCount));
        for (U16 i = 0; i < basicTypeArraysWithCount_CountM; i++) {
            basicTypeArraysWithCount[i] = new BasicTypeArraysWithCount;
            basicTypeArraysWithCount[i]->Randomize();
        }

        Rand<U32>(basicTypePointersWithMax_CountM,
                  Util__CountOf(basicTypePointersWithMax));
        for (U32 i = 0; i < basicTypePointersWithMax_CountM; i++) {
            basicTypePointersWithMax[i] = new BasicTypePointersWithMax;
            basicTypePointersWithMax[i]->Randomize();
        }

        Rand<U32>(basicTypePointersWithCountAndMax_CountM,
                  Util__CountOf(basicTypePointersWithCountAndMax));
        for (U32 i = 0; i < basicTypePointersWithCountAndMax_CountM;
             i++) {
            basicTypePointersWithCountAndMax[i] =
                new LocalBasicTypePointersWithCountAndMax;
            basicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

private:
    
    S8 basicTypes_CountM;
    LocalBasicTypes *basicTypes[4];

    U8 basicTypeArrays_CountM;
    BasicTypeArrays *basicTypeArrays[6];

    U16 basicTypeArraysWithCount_CountM;
    BasicTypeArraysWithCount *basicTypeArraysWithCount[10];

    U32 basicTypePointersWithMax_CountM;
    BasicTypePointersWithMax *basicTypePointersWithMax[6];

    U32 basicTypePointersWithCountAndMax_CountM;
    LocalBasicTypePointersWithCountAndMax *
        basicTypePointersWithCountAndMax[4];
};

class LocalStructDoublePointersWithMax : public LocalTestClass
{
public:

    inline LocalStructDoublePointersWithMax()
        : pBasicTypes(NULL),
          pBasicTypeArrays(NULL),
          pBasicTypeArraysWithCount(NULL),
          pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax(NULL)
    {
    }

    virtual ~LocalStructDoublePointersWithMax()
    {
        if (pBasicTypes) {
            for (int i = 0; i < pBasicTypes_Max; i++) {
                delete pBasicTypes[i];
            }
            delete [] pBasicTypes;
        }

        if (pBasicTypeArrays) {
            for (int i = 0; i < pBasicTypeArrays_Max; i++) {
                delete pBasicTypeArrays[i];
            }
            delete [] pBasicTypeArrays;
        }

        if (pBasicTypeArraysWithCount) {
            for (int i = 0; i < pBasicTypeArraysWithCount_Max; i++) {
                delete pBasicTypeArraysWithCount[i];
            }
            delete [] pBasicTypeArraysWithCount;
        }

        if (pBasicTypePointersWithMax) {
            for (int i = 0; i < pBasicTypePointersWithMax_Max; i++) {
                delete pBasicTypePointersWithMax[i];
            }
            delete [] pBasicTypePointersWithMax;
        }

        if (pBasicTypePointersWithCountAndMax) {
            for (int i = 0; i < pBasicTypePointersWithCountAndMax_Max;
                 i++) {
                delete pBasicTypePointersWithCountAndMax[i];
            }
            delete [] pBasicTypePointersWithCountAndMax;
        }
    }

    virtual void Randomize()
    {
        pBasicTypes = new BasicTypes * [pBasicTypes_Max];
        for (int i = 0; i < pBasicTypes_Max; i++) {
            pBasicTypes[i] = new BasicTypes;
            pBasicTypes[i]->Randomize();
        }

        pBasicTypeArrays = new LocalBasicTypeArrays *
            [pBasicTypeArrays_Max];
        for (int i = 0; i < pBasicTypeArrays_Max; i++) {
            pBasicTypeArrays[i] = new LocalBasicTypeArrays;
            pBasicTypeArrays[i]->Randomize();
        }

        pBasicTypeArraysWithCount = new LocalBasicTypeArraysWithCount *
            [pBasicTypeArraysWithCount_Max];
        for (int i = 0; i < pBasicTypeArraysWithCount_Max; i++) {
            pBasicTypeArraysWithCount[i] = new
                LocalBasicTypeArraysWithCount;
            pBasicTypeArraysWithCount[i]->Randomize();
        }

        pBasicTypePointersWithMax = 
            new LocalBasicTypePointersWithMax *
            [pBasicTypePointersWithMax_Max];
        for (int i = 0; i < pBasicTypePointersWithMax_Max; i++) {
            pBasicTypePointersWithMax[i] = new
                LocalBasicTypePointersWithMax;
            pBasicTypePointersWithMax[i]->Randomize();
        }

        pBasicTypePointersWithCountAndMax = 
            new BasicTypePointersWithCountAndMax *
            [pBasicTypePointersWithCountAndMax_Max];
        for (int i = 0; i < pBasicTypePointersWithCountAndMax_Max;
             i++) {
            pBasicTypePointersWithCountAndMax[i] = new
                BasicTypePointersWithCountAndMax;
            pBasicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

private:
    
    static const int pBasicTypes_Max = 4;
    BasicTypes **pBasicTypes;

    static const int pBasicTypeArrays_Max = 6;
    LocalBasicTypeArrays **pBasicTypeArrays;

    static const int pBasicTypeArraysWithCount_Max = 10;
    LocalBasicTypeArraysWithCount **pBasicTypeArraysWithCount;

    static const int pBasicTypePointersWithMax_Max = 6;
    LocalBasicTypePointersWithMax **pBasicTypePointersWithMax;

    static const int pBasicTypePointersWithCountAndMax_Max = 4;
    BasicTypePointersWithCountAndMax **pBasicTypePointersWithCountAndMax;
};

class LocalStructDoublePointersWithCountAndMax : public LocalTestClass
{
public:

    inline LocalStructDoublePointersWithCountAndMax()
        : pBasicTypes_CountM(0),
          pBasicTypes(NULL),
          pBasicTypeArrays_CountM(0),
          pBasicTypeArrays(NULL),
          pBasicTypeArraysWithCount_CountM(0),
          pBasicTypeArraysWithCount(NULL),
          pBasicTypePointersWithMax_CountM(0),
          pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax_CountM(0),
          pBasicTypePointersWithCountAndMax(NULL)
    {
    }

    virtual ~LocalStructDoublePointersWithCountAndMax()
    {
        for (S8 i = 0; i < pBasicTypes_CountM; i++) {
            delete pBasicTypes[i];
        }
        delete [] pBasicTypes;

        for (U8 i = 0; i < pBasicTypeArrays_CountM; i++) {
            delete pBasicTypeArrays[i];
        }
        delete [] pBasicTypeArrays;

        for (U16 i = 0; i < pBasicTypeArraysWithCount_CountM; i++) {
            delete pBasicTypeArraysWithCount[i];
        }
        delete [] pBasicTypeArraysWithCount;

        for (U32 i = 0; i < pBasicTypePointersWithMax_CountM; i++) {
            delete pBasicTypePointersWithMax[i];
        }
        delete [] pBasicTypePointersWithMax;

        for (U32 i = 0; i < pBasicTypePointersWithCountAndMax_CountM;
             i++) {
            delete pBasicTypePointersWithCountAndMax[i];
        }
        delete [] pBasicTypePointersWithCountAndMax;
    }

    virtual void Randomize()
    {
        Rand<S8>(pBasicTypes_CountM, pBasicTypes_Max);
        pBasicTypes = new LocalBasicTypes * [pBasicTypes_CountM];
        for (S8 i = 0; i < pBasicTypes_CountM; i++) {
            pBasicTypes[i] = new LocalBasicTypes;
            pBasicTypes[i]->Randomize();
        }

        Rand<U8>(pBasicTypeArrays_CountM, pBasicTypeArrays_Max);
        pBasicTypeArrays = new BasicTypeArrays * 
            [pBasicTypeArrays_CountM];
        for (U8 i = 0; i < pBasicTypeArrays_CountM; i++) {
            pBasicTypeArrays[i] = new BasicTypeArrays;
            pBasicTypeArrays[i]->Randomize();
        }

        Rand<U16>(pBasicTypeArraysWithCount_CountM,
                  pBasicTypeArraysWithCount_Max);
        pBasicTypeArraysWithCount = new BasicTypeArraysWithCount * 
            [pBasicTypeArraysWithCount_CountM];
        for (U16 i = 0; i < pBasicTypeArraysWithCount_CountM; i++) {
            pBasicTypeArraysWithCount[i] = 
                new BasicTypeArraysWithCount;
            pBasicTypeArraysWithCount[i]->Randomize();
        }

        Rand<U32>(pBasicTypePointersWithMax_CountM,
                  pBasicTypePointersWithMax_Max);
        pBasicTypePointersWithMax = new LocalBasicTypePointersWithMax * 
            [pBasicTypePointersWithMax_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithMax_CountM; i++) {
            pBasicTypePointersWithMax[i] =
                new LocalBasicTypePointersWithMax;
            pBasicTypePointersWithMax[i]->Randomize();
        }

        Rand<U32>(pBasicTypePointersWithCountAndMax_CountM,
                  pBasicTypePointersWithCountAndMax_Max);
        pBasicTypePointersWithCountAndMax = 
            new BasicTypePointersWithCountAndMax * 
            [pBasicTypePointersWithCountAndMax_CountM];
        for (U32 i = 0; i < pBasicTypePointersWithCountAndMax_CountM;
             i++) {
            pBasicTypePointersWithCountAndMax[i] =
                new BasicTypePointersWithCountAndMax;
            pBasicTypePointersWithCountAndMax[i]->Randomize();
        }
    }

private:
    
    static const int pBasicTypes_Max = 4;
    S8 pBasicTypes_CountM;
    LocalBasicTypes **pBasicTypes;

    static const int pBasicTypeArrays_Max = 6;
    U8 pBasicTypeArrays_CountM;
    BasicTypeArrays **pBasicTypeArrays;

    static const int pBasicTypeArraysWithCount_Max = 10;
    U16 pBasicTypeArraysWithCount_CountM;
    BasicTypeArraysWithCount **pBasicTypeArraysWithCount;

    static const int pBasicTypePointersWithMax_Max = 6;
    U32 pBasicTypePointersWithMax_CountM;
    LocalBasicTypePointersWithMax **pBasicTypePointersWithMax;

    static const int pBasicTypePointersWithCountAndMax_Max = 4;
    U32 pBasicTypePointersWithCountAndMax_CountM;
    BasicTypePointersWithCountAndMax **pBasicTypePointersWithCountAndMax;
};

class LocalDeepInnerBasicType : public LocalTestClass
{
public:

    virtual void Randomize()
    {
        Rand<U64>(outer.middle.inner.u64);
    }

private:

    struct {
        typedef struct {
            int a;
        } wonky;
        wonky w;
        class {
        public:
            struct {
                U64 u64;
            } inner;
        } middle;
    } outer;
};

class LocalDeepInnerStructAndHierarchy : public LocalTestClass
{
public:

    virtual void Randomize()
    {
        outer.middle.Randomize();
    }

private:

    struct {
        class {
        public:

            void Randomize()
            {
                inner.upper.Randomize();
                inner.middle.Randomize();
                inner.bottom.Randomize();
            }

            struct {

                class Upper
                {
                public:

                    void Randomize()
                    {
                        upperOuter.Randomize();
                    }

                    class UpperOuter
                    {
                    public:

                        void Randomize()
                        {
                            serialized.middle.Randomize();

                            serialized.bottom.Randomize();

                            RandReal<double>(transient_double);
                        }

                        class Middle
                        {
                        public:

                            void Randomize()
                            {
                                Rand<U8>(u8);
                            }
                            
                        private:

                            U8 u8;
                        };

                        class Bottom : public Middle
                        {
                        public:

                            void Randomize()
                            {
                                RandReal<float>(serialized.fl);
                            }

                        private:

                            struct {
                                float fl;
                            } serialized;
                        };

                        struct {
                            Middle middle;
                            Bottom bottom;
                        } serialized;

                        double transient_double;
                    };

                    UpperOuter upperOuter;
                };

                class Middle : public Upper
                {
                public:

                    void Randomize()
                    {
                        Rand<U16>(serialized.u16);
                        
                        RandReal<double>(transient_double);
                    }

                private:
                    
                    struct {
                        U16 u16;
                    } serialized;

                    double transient_double;
                };
                
                class Bottom : public Middle
                {
                public:

                    void Randomize()
                    {
                        Rand<U32>(serialized.u32);

                        RandReal<double>(transient_double);
                    }

                private:
                    
                    struct {
                        U32 u32;
                    } serialized;

                    double transient_double;
                };
                
                Upper upper;
                Middle middle;
                Bottom bottom;
            } inner;
        } middle;
    } outer;
};

class LocalBigArray : public LocalTestClass
{
public:

    virtual void Randomize()
    {
        U32 array_count;
        Rand<U32>(array_count, Util__CountOf(array) - 1);

        for (U32 i = 0; i < array_count; i++) {
            array[i] = (Util::RandomNumber() % ('z' - 'a')) + 'a';
        }
        array[array_count] = 0;

        for (U32 i = 0; i < Util__CountOf(array2); i++) {
            RandReal<float>(array2[i]);
        }
    }

private:

    U8 array[1005];

    float array2[1020];
};

class LocalSimpleStructArray : public LocalTestClass
{
public:
    
    virtual void Randomize()
    {
        Rand<U32>(simpleStructs_CountM, Util__CountOf(simpleStructs));

        for (U32 i = 0; i < simpleStructs_CountM; i++) {
            Rand<U32>(simpleStructs[i].val1);
            Rand<U32>(simpleStructs[i].val2);
            Rand<U32>(simpleStructs[i].val3);
        }
    }

private:

    class SimpleStruct
    {
    public:

        U32 val1, val2, val3;
    };

    U32 simpleStructs_CountM;
    SimpleStruct simpleStructs[100];
};

class LocalOptional : public LocalTestClass
{
public:
    
    LocalOptional()
        : pBasicTypes(NULL), pBasicTypeArrays(NULL), 
          pBasicTypeArraysWithCount(NULL), pBasicTypePointersWithMax(NULL),
          pBasicTypePointersWithCountAndMax(NULL), pStructs(NULL),
          pStructArrays(NULL), pStructPointers(NULL), 
          pStructArraysWithCount(NULL), pStructPointersWithCountAndMax(NULL), 
          pStructArrayOfPointers(NULL), pStructArrayOfPointersWithCount(NULL), 
          pStructDoublePointersWithMax(NULL), 
          pStructDoublePointersWithCountAndMax(NULL),
          pDeepInnerBasicType(NULL), pDeepInnerStructAndHierarchy(NULL), 
          pBigArray(NULL), pSimpleStructArray(NULL)
    {
    }

    virtual ~LocalOptional()
    {
        delete pBasicTypes;
        delete pBasicTypeArrays;
        delete pBasicTypeArraysWithCount;
        delete pBasicTypePointersWithMax;
        delete pBasicTypePointersWithCountAndMax;
        delete pStructs;
        delete pStructArrays;
        delete pStructPointers;
        delete pStructArraysWithCount;
        delete pStructPointersWithCountAndMax;
        delete pStructArrayOfPointers;
        delete pStructArrayOfPointersWithCount;
        delete pStructDoublePointersWithMax;
        delete pStructDoublePointersWithCountAndMax;
        delete pDeepInnerBasicType;
        delete pDeepInnerStructAndHierarchy;
        delete pBigArray;
        delete pSimpleStructArray;
    }

    virtual void Randomize()
    {
        if (Util::RandomNumber() % 2) {
            pBasicTypes = new LocalBasicTypes();
            pBasicTypes->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArrays = new LocalBasicTypeArrays();
            pBasicTypeArrays->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypeArraysWithCount = new LocalBasicTypeArraysWithCount();
            pBasicTypeArraysWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithMax = new LocalBasicTypePointersWithMax();
            pBasicTypePointersWithMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBasicTypePointersWithCountAndMax = new
                LocalBasicTypePointersWithCountAndMax();
            pBasicTypePointersWithCountAndMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructs = new LocalStructs();
            pStructs->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArrays = new LocalStructArrays();
            pStructArrays->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructPointers = new LocalStructPointers();
            pStructPointers->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArraysWithCount = new LocalStructArraysWithCount();
            pStructArraysWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructPointersWithCountAndMax = new
                StructPointersWithCountAndMax();
            pStructPointersWithCountAndMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArrayOfPointers = new LocalStructArrayOfPointers();
            pStructArrayOfPointers->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructArrayOfPointersWithCount = new
                StructArrayOfPointersWithCount();
            pStructArrayOfPointersWithCount->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructDoublePointersWithMax = 
                new LocalStructDoublePointersWithMax();
            pStructDoublePointersWithMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pStructDoublePointersWithCountAndMax = new
                StructDoublePointersWithCountAndMax();
            pStructDoublePointersWithCountAndMax->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pDeepInnerBasicType = new LocalDeepInnerBasicType();
            pDeepInnerBasicType->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pDeepInnerStructAndHierarchy = new DeepInnerStructAndHierarchy();
            pDeepInnerStructAndHierarchy->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pBigArray = new LocalBigArray();
            pBigArray->Randomize();
        }
        if (Util::RandomNumber() % 2) {
            pSimpleStructArray = new SimpleStructArray();
            pSimpleStructArray->Randomize();
        }
    }

private:

    LocalBasicTypes *pBasicTypes;
    LocalBasicTypeArrays *pBasicTypeArrays;
    LocalBasicTypeArraysWithCount *pBasicTypeArraysWithCount;
    LocalBasicTypePointersWithMax *pBasicTypePointersWithMax;
    LocalBasicTypePointersWithCountAndMax *pBasicTypePointersWithCountAndMax;
    LocalStructs *pStructs;
    LocalStructArrays *pStructArrays;
    LocalStructPointers *pStructPointers;
    LocalStructArraysWithCount *pStructArraysWithCount;
    StructPointersWithCountAndMax *pStructPointersWithCountAndMax;
    LocalStructArrayOfPointers *pStructArrayOfPointers;
    StructArrayOfPointersWithCount *pStructArrayOfPointersWithCount;
    LocalStructDoublePointersWithMax *pStructDoublePointersWithMax;
    StructDoublePointersWithCountAndMax *pStructDoublePointersWithCountAndMax;
    LocalDeepInnerBasicType *pDeepInnerBasicType;
    DeepInnerStructAndHierarchy *pDeepInnerStructAndHierarchy;
    LocalBigArray *pBigArray;
    SimpleStructArray *pSimpleStructArray;
};

class LocalTester : public TestClass
{
public:

    virtual void Randomize()
    {
        localOptionalM.Randomize();
    }

    virtual const char *GetIdentity() const 
    { 
        return "LocalTester";
    }

private:

    LocalOptional localOptionalM;

};

/* static */
inline TestClass *TestClass::Create(bool fixedId, U32 id)
{
    TestClass *pClass;

    if (!fixedId) {
        id = Util::RandomNumber() % 21;
    }

    switch (id) {
    case 0:
        pClass = new BasicTypes();
        break;
    case 1:
        pClass = new BasicTypeArrays();
        break;
    case 2:
        pClass = new BasicTypeArraysWithCount();
        break;
    case 3:
        pClass = new BasicTypePointersWithCount();
        break;
    case 4:
        pClass = new BasicTypePointersWithMax();
        break;
    case 5:
        pClass = new BasicTypePointersWithCountAndMax();
        break;
    case 6:
        pClass = new Structs();
        break;
    case 7:
        pClass = new StructArrays();
        break;
    case 8:
        pClass = new StructPointers();
        break;
    case 9:
        pClass = new StructArraysWithCount();
        break;
    case 10:
        pClass = new StructPointersWithCountAndMax();
        break;
    case 11:
        pClass = new StructArrayOfPointers();
        break;
    case 12:
        pClass = new StructArrayOfPointersWithCount();
        break;
    case 13:
        pClass = new StructDoublePointersWithMax();
        break;
    case 14:
        pClass = new StructDoublePointersWithCountAndMax();
        break;
    case 15:
        pClass = new DeepInnerBasicType();
        break;
    case 16:
        pClass = new DeepInnerStructAndHierarchy();
        break;
    case 17:
        pClass = new BigArray();
        break;
    case 18:
        pClass = new SimpleStructArray();
        break;
    case 19:
        pClass = new Optional();
        break;
    case 20:
        pClass = new LocalTester();
        break;
    default:
        Log::ErrorF("Unknown id: %lu\n", id);
        Util::Terminate();
    }

    pClass->Randomize();

    return pClass;
}

#endif // TEST_CLASSES_H
