/*****************************************************************************\
 *                                                                           *
 * Parsed.cpp                                                                 *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include "private/Parsed.h"

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif

static String g_Anonymous = "<anonymous>";

/* static */
void split(const unichar *pString, RefVector<RefString> &v)
{
    const unichar *p = pString;
    while (true) {
        // Skip whitespace to new member
        while (*p && (*p == ' ')) {
            p++;
        }
        if (!*p) {
            break;
        }
        // This is the start of the member
        pString = p;
        // Skip to whitespace
        while (*p && (*p != ' ')) {
            p++;
        }
        // Now add the id
        ref<RefString> pValue = new RefString(pString, 0, p - pString);
        v.Append(pValue);
    }
}


/* static */
char *get_name(const Parser::Element *pElement, bool anonymous)
{
    String name = pElement->GetAttributeValue("name");
    const unichar *u_name;

    // gccxml apparently has decided to forego names on some contexts now.
    // The name can be recovered from the "demangled" field though.
    if (name == "") {
        String demangled = pElement->GetAttributeValue("demangled");
        if (demangled == "") {
            return String::DuplicateChars(anonymous ? g_Anonymous : "");
        }
        else {
            // Find the last colon in demangled, and use the rest as name
            const unichar *demangled_u = (const unichar *) demangled;
            const unichar *end = &(demangled_u[demangled.Length() - 1]);
            while ((end > demangled_u) && (*end != ':')) {
                end--;
            }
            if (*end == ':') {
                end++;
            }
            u_name = end;
        }
    }
    else {
        u_name = (const unichar *) name;
    }

    // If it's a gccxml anonymous name, use <anonymous>
    if (anonymous && (u_name[0] == '.')) {
        return String::DuplicateChars(g_Anonymous);
    }
    
    return String::DuplicateChars(u_name);
}


/* static */
AccessType get_access_type(const Parser::Element *pElement)
{
    String access = pElement->GetAttributeValue("access");

    if (access == "private") {
        return AccessType_Private;
    }
    else if (access == "protected") {
        return AccessType_Protected;
    }
    else {
        return AccessType_Public;
    }
}


/* static */
bool get_member_elements(Parser &parser,const Parser::Element *pElement, 
                         RefVector<Parser::Element> &vMembers, String &error)
{
    String memberIds = pElement->GetAttributeValue("members");

    RefVector<RefString> vMemberIds;

    split((const unichar *) memberIds, vMemberIds);

    U32 membersCount = vMemberIds.Count();

    for (U32 i = 0; i < membersCount; i++) {
        String memberId = String(vMemberIds[i]);
        ref<Parser::Element> pMemberElement = parser.LookupElement(memberId);

        if (!pMemberElement) {
            error = String::Format
                ("No such member: %S", (const unichar *) memberId);
            return false;
        }

        // Skip artifical members
        if (pMemberElement->GetAttributeValue("artificial") == "1") {
            continue;
        }

        vMembers.Append(pMemberElement);
    }
    
    return true;
}


ContextSet *CreateContextSet()
{
    return new ParsedContextSet();
}


}; // namespace Xrtti
