/*****************************************************************************\
 *                                                                           *
 * ParsedStructure.cpp                                                       *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include "private/Parsed.h"

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


ParsedStructure::ParsedStructure()
    : pBasesM(0), pFieldsM(0), pConstructorsM(0), pDestructorM(0)
{
}


ParsedStructure::~ParsedStructure()
{
    delete [] pBasesM;
    delete [] pFieldsM;
    delete [] pConstructorsM;
    delete pDestructorM;
}

  
bool ParsedStructure::Initialize(ParsedContextSet &set, Parser &parser,
                                 const Parser::Element *pElement, String &error)
{
    // ParsedContext
    if (!superM.Initialize(set, parser, pElement, error)) {
        return false;;
    }

    // Get a pointer to this context - this is done because this context
    // actually may not be a ParsedStructure; we cannot use our "this" pointer
    // as a pointer to this context, because we may just be imbedded in
    // another object (ParsedStruct or ParsedUnion)
    Context *pThisContext = set.GetContext
        (parser, pElement->GetAttributeValue("id"), error);

    // isAnonymousM must be done out-of-order: it must be done early, so
    // that it can be initialized before any subclasses end up being
    // initialized (subclasses are initialized when loaded below in this
    // initializer, and isAnonymousM has to be initialized before then)
    // bool isAnonymousM;
    String name = this->GetName();
    if (name == "<anonymous>") {
        isAnonymousM = true;
    }
    else {
        const Context *pContext = this->GetContext();

        if (pContext) {
            switch (pContext->GetType()) {
            case Type_Class:
            case Type_Struct:
            case Type_Union:
                isAnonymousM = ((const Structure *) pContext)->IsAnonymous();
                break;
            default: // Type_Namespace
                isAnonymousM = false;
                break;
            }
        }
        else {
            isAnonymousM = false;
        }
    }

    // Friends
    String befriending = pElement->GetAttributeValue("befriending");
    RefVector<RefString> vBefriendings;
    split((const unichar *) befriending, vBefriendings);

    U32 count = vBefriendings.Count();
    for (U32 i = 0; i < count; i++) {
        Context *pContext = set.GetContext
            (parser, String(vBefriendings[i]), error);
        if (pContext == NULL) {
            return false;
        }
        switch (pContext->GetType()) {
        case Context::Type_Class:
            ((ParsedClass *) pContext)->
                AddFriend((Structure *) pThisContext);
            break;
        case Context::Type_Struct:
            ((ParsedStruct *) pContext)->
                AddFriend((Structure *) pThisContext);
            break;
        default:
            error = String::Format
                ("Unexpected context type %lu for friend %lu of structure %S",
                 pContext->GetType(), i, (const unichar *)
                 pElement->GetAttributeValue("id"));
            return false;
        }
    }

    // bool isIncompleteM
    isIncompleteM = (pElement->GetAttributeValue("incomplete") == "1");
    if (isIncompleteM) {
        // just in case gccxml tries to be funny and give contents for
        // an incomplete structure, ignore anything more
        accessTypeM = AccessType_Public;
        baseCountM = fieldCountM = constructorCountM = 0;
        pDestructorM = 0;
        return true;
    }

    // bool hasStructureNameM
    hasStructureNameM = (pElement->GetAttributeValue("artificial") == "1");

    // AccessType accessTypeM;
    accessTypeM = get_access_type(pElement);
        
    // U32 baseCountM;
    baseCountM = 0;
    count = pElement->GetSubElementCount();
    for (U32 i = 0; i < count; i++) {
        ref<Parser::Element> pSubElement = pElement->GetSubElement(i);
        if (pSubElement->GetElementName() == "Base") {
            baseCountM++;
        }
    }

    // Base *pBasesM;
    if (baseCountM) {
        pBasesM = new ParsedBase[baseCountM];

        U32 which = 0;
        for (U32 i = 0; i < count; i++) {
            ref<Parser::Element> pSubElement = pElement->GetSubElement(i);
            if (pSubElement->GetElementName() == "Base") {
                if (!pBasesM[which++].Initialize
                    (set, parser, pSubElement->GetAttributeValue("type"), 
                     get_access_type(pSubElement),
                     pSubElement->GetAttributeValue("virtual") == "1",
                     error)) {
                    return false;
                }
            }
        }
    }

    // Get the members elements
    RefVector<Parser::Element> vMemberElements;

    if (!get_member_elements(parser, pElement, vMemberElements, error)) {
        return false;;
    }

    // U32 fieldCountM
    // U32 constructorCountM
    fieldCountM = 0, constructorCountM = 0;
    count = vMemberElements.Count();
    for (U32 i = 0; i < count; i++) {
        ref<Parser::Element> pMemberElement = vMemberElements[i];

        String memberElementName = pMemberElement->GetElementName();

        // If this member is a Field, count it as a field
        if (memberElementName == "Field") {
            fieldCountM++;
        }
        // If this member is a Variable, count add it as a static field
        else if (memberElementName == "Variable") {
            // Skip it if it's got no name, which gccxml does for no obvious
            // reason sometimes
            if (pMemberElement->GetAttributeValue("name") == "") {
                continue;
            }
            fieldCountM++;
        }
        // Else if it's a Constructor, then count it as a constructor
        else if (memberElementName == "Constructor") {
            constructorCountM++;
        }
        // Else ignore it
    }

    // Field *pFieldsM;
    if (fieldCountM) {
        pFieldsM = new ParsedField[fieldCountM];
    }
        
    // Constructor *pConstructorsM;
    if (constructorCountM) {
        pConstructorsM = new ParsedConstructor[constructorCountM];
    }

    // Go over all members, collecting fields, constructors, and the
    // destructor
    U32 fieldIndex = 0, constructorIndex = 0;

    // Destructor *pDestructorM;
    for (U32 i = 0; i < count; i++) {
        ref<Parser::Element> pMemberElement = vMemberElements[i];

        String memberElementName = pMemberElement->GetElementName();

        // If this member is a Field, then initialize it
        if (memberElementName == "Field") {
            if (!pFieldsM[fieldIndex++].Initialize
                (set, parser, pThisContext, false, pMemberElement, error)) {
                return false;;
            }
        }
        // If this member is a Variable, then initialize it as a static
        else if (memberElementName == "Variable") {
            // Skip it if it's got no name, which gccxml does for no obvious
            // reason sometimes
            if (pMemberElement->GetAttributeValue("name") = "") {
                continue;
            }
            if (!pFieldsM[fieldIndex++].Initialize
                (set, parser, pThisContext, true, pMemberElement, error)) {
                return false;;
            }
        }
        // Else if it's a Constructor, then initialize it
        else if (memberElementName == "Constructor") {
            if (!pConstructorsM[constructorIndex++].Initialize
                (set, parser, pThisContext, pMemberElement, error)) {
                return false;;
            }
        }
        // Else if it's a Destructor, then initialize it
        else if (memberElementName == "Destructor") {
            if (pDestructorM) {
                error = String::Format
                    ("Duplicate destructor in structure id %S",
                     (const unichar *) pElement->GetAttributeValue("id"));
                return false;;
            }
            pDestructorM = new ParsedDestructor();
            if (!pDestructorM->Initialize
                (set, parser, pThisContext, pMemberElement, error)) {
                return false;;
            }
        }
        // Otherwise, ignore it, we don't care about it
    }
        
    return true;
}


void ParsedStructure::MergeContents(ParsedContextSet &to, 
                                    ParsedContextSet &from)
{
    superM.MergeContents(to, from);

    for (U32 i = 0; i < baseCountM; i++) {
        pBasesM[i].MergeContents(to, from);
    }

    for (U32 i = 0; i < fieldCountM; i++) {
        pFieldsM[i].MergeContents(to, from);
    }

    for (U32 i = 0; i < constructorCountM; i++) {
        pConstructorsM[i].MergeContents(to, from);
    }

    if (pDestructorM) {
        pDestructorM->MergeContents(to, from);
    }
}


void ParsedStructure::ReplaceContext(Context *pOld, Context *pNew)
{
    U32 count = vFriendsM.Count();
    for (U32 i = 0; i < count; i++) {
        if (vFriendsM[i] == pOld) {
            vFriendsM[i] = (Structure *) pNew;
        }
    }

    for (U32 i = 0; i < fieldCountM; i++) {
        pFieldsM[i].ReplaceContext(pOld, pNew);
    }

    for (U32 i = 0; i < constructorCountM; i++) {
        pConstructorsM[i].ReplaceContext(pOld, pNew);
    }
}


const Base &ParsedStructure::GetBase(U32 index) const
{
    return pBasesM[index];
}


const Structure &ParsedStructure::GetFriend(U32 index) const
{
    return *(vFriendsM[index]);
}


const Field &ParsedStructure::GetField(U32 index) const
{
    return pFieldsM[index];
}


const Constructor &ParsedStructure::GetConstructor(U32 index) const
{
    return pConstructorsM[index];
}


const Destructor &ParsedStructure::GetDestructor() const
{
    return *pDestructorM;
}


}; // namespace Xrtti
