/*****************************************************************************\
 *                                                                           *
 * ParsedType.cpp                                                            *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include "private/Parsed.h"

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


// This helper type holds keeps track of Type attributes
typedef struct Attributes
{
    bool is_const;
    bool is_volatile;
} Attributes;

// This helper type holds keeps track of Type arrays
typedef struct ArrayOrPointerProps
{
    bool is_array;

    // array props
    bool is_unbounded;
    U32 max;

    // pointer props
    bool is_const;
    bool is_volatile;
} ArrayOrPointerProps;



ParsedType::ParsedType()
    : pArrayOrPointersM(0)
{
}

ParsedType::~ParsedType()
{
    for (U32 i = 0; i < arrayOrPointerCountM; i++) {
        delete pArrayOrPointersM[i];
    }
    delete [] pArrayOrPointersM;
}

bool ParsedType::Initialize(ParsedContextSet &set, Parser &parser,
                            Type::BaseType baseType, Parser::Element *pElement,
                            String &error)
{
    baseTypeM = baseType;

    // bool isReferenceM;
    if (pElement->GetElementName() == "ReferenceType") {
        // Set reference to true
        isReferenceM = true;
        // Get the type that is referenced
        if (!(pElement = parser.LookupElement
              (pElement->GetAttributeValue("type")))) {
            error = String::Format
                ("No such type [%S] for reference type %S",
                 (const unichar *) pElement->GetAttributeValue("type"),
                 (const unichar *) pElement->GetAttributeValue("id"));
            return false;
        }
    }
    else {
        isReferenceM = false;
    }

    // Now loop through the remaining types, accumulating pointer
    // and array info
    Vector<ArrayOrPointerProps> vArrayOrPointerProps;
    Attributes attributes = { false, false };
    while (true) {
        String type = pElement->GetElementName();

        // CvQualifiedType simply adds qualifying attributes to the next
        // type specifier
        if (type == "CvQualifiedType") {
            attributes.is_const = 
                (pElement->GetAttributeValue("const") == "1");
            attributes.is_volatile =
                (pElement->GetAttributeValue("volatile") == "1");
            // Ignore restrict, it is g++-specific
        }
        // PointerType adds to the pointer level vector, with the last
        // used attributes
        else if (type == "PointerType") {
            ArrayOrPointerProps props = { false, false, 0, attributes.is_const, 
                                          attributes.is_volatile };
            vArrayOrPointerProps.Append(props);
            attributes.is_const = attributes.is_volatile = false;
        }
        // ArrayType adds another array dimension, with its max
        else if (type == "ArrayType") {
            ArrayOrPointerProps props = { true, false, 0, false, false };
            String max = pElement->GetAttributeValue("max");
            if ((max == "ffffffffffffffff") ||
                (max == "0xffffffffffffffff")) {
                // This indicates no practical maximumum
                props.is_unbounded = true;
            }
            else {
                props.max = max.ToU64();
            }

            vArrayOrPointerProps.Append(props);
        }
        // Skip right over typedefs, to the actual type
        else if (type == "Typedef") {
        }
        // Other types finish the type definition
        else {
            break;
        }
        
        // Next sub type
        if (!(pElement = parser.LookupElement
              (pElement->GetAttributeValue("type")))) {
            error = String::Format
                ("Expected a sub type for type id %S",
                 (const unichar *) pElement->GetAttributeValue("id"));
            return false;
        }
    }

    // bool isConstM;
    isConstM = attributes.is_const;

    // bool isVolatileM;
    isVolatileM = attributes.is_volatile;

    // U32 arrayOrPointerCountM;
    arrayOrPointerCountM = vArrayOrPointerProps.Count();

    // Pointer *pPointersM;
    if (arrayOrPointerCountM) {
        pArrayOrPointersM = new ArrayOrPointer *[arrayOrPointerCountM];
        for (U32 i = 0; i < arrayOrPointerCountM; i++) {
            ArrayOrPointerProps props = vArrayOrPointerProps[i];
            if (props.is_array) {
                pArrayOrPointersM[i] = new ParsedArray
                    (props.is_unbounded, props.max + 1);
            }
            else {
                pArrayOrPointersM[i] = new ParsedPointer
                    (props.is_const, props.is_volatile);
            }
        }
    }

    return true;
}


const ArrayOrPointer &ParsedType::GetArrayOrPointer(U32 index) const
{
    return *(pArrayOrPointersM[index]);
}


}; // namespace Xrtti
