/*****************************************************************************\
 *                                                                           *
 * Parser.cpp                                                                *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include <Essential/Log.h>
#include "../../inc/private/Parser.h"


// --------------------------------------------------------------------------
// Parser::Parser(const Configuration &config, InputStream &instream)
// --------------------------------------------------------------------------
Parser::Parser()
    : parseErrorM(false), xmlParserM(NULL)
{
}


// --------------------------------------------------------------------------
// Parser::Parse(InputStream &instream)
// --------------------------------------------------------------------------
void Parser::Parse(InputStream &instream)
{
    // Set the expat parser
    xmlParserM = XML_ParserCreate("utf-8");
    XML_SetStartElementHandler(xmlParserM, ExpatStartElementHandler);
    XML_SetEndElementHandler(xmlParserM, ExpatEndElementHandler);
    XML_SetUserData(xmlParserM, this);
    
    while (true) {
        U8 inbuf[1024];
        U32 bytesRead;

        bool fReadSuccess = instream.ReadBytes
            (Util__CountOf(inbuf), inbuf, /* returns */ bytesRead);
        
        if (!fReadSuccess) {
            if (!XML_Parse(xmlParserM, (const char *) inbuf, 0, true) || 
                parseErrorM) {
                this->XmlError("XML parse failed\n");
            }
            break;
        }

        if (!XML_Parse(xmlParserM, (const char *) inbuf, bytesRead, false) || 
            parseErrorM) {
            this->XmlError("XML parse failed\n");
        }
    }

    XML_ParserFree(xmlParserM);

    xmlParserM = NULL;
}


// --------------------------------------------------------------------------
// Parser::ExpatStartElementHandler(void *pData,
//                                  const XML_Char *pName,
//                                  const XML_Char **pAttributes)
// --------------------------------------------------------------------------
/* static */
void Parser::ExpatStartElementHandler(void *pData,
                                      const XML_Char *pName,
                                      const XML_Char **pAttributes)
{
    ((Parser *) pData)->StartElement((char *) pName, (char **) pAttributes);
}


// --------------------------------------------------------------------------
// Parser::ExpatEndElementHandler(void *pData, const XML_Char *pName)
// --------------------------------------------------------------------------
/* static */
void Parser::ExpatEndElementHandler(void *pData, const XML_Char *pName)
{
    ((Parser *) pData)->EndElement((char *) pName);
}


// --------------------------------------------------------------------------
// Parser::StartElement(char *pElementName, char **pAttributes)
// --------------------------------------------------------------------------
void Parser::StartElement(char *pElementName, char **pAttributes)
{
    // Ignore GCC-XML element
    if (String::Equal(pElementName, "GCC_XML")) {
        return;
    }

    // Create a new Element for this XML element
    ref<Element> pElement = new Element(*this, pElementName, pAttributes);

    // See if this XML element is a child of another XML element, which
    // will already have been "pushed" on the current elements "stack"
    ref<Element> pSuper = vCurrentElementsM.Last();

    // If there is no super element, then this is a "root" element
    if (pSuper == NULL) {
        // root elements are expected to have ids
        String id = pElement->GetAttributeValue("id");
        if (id == "") {
            this->XmlError("missing id");
        }
        htElementsM.Put(id, pElement);
    }
    // If there is a super element, then add this element to it
    else {
        pSuper->subElementsM.Append(pElement);
    }
             
    // "Push" the current element on the element "stack"
    vCurrentElementsM.Append(pElement);
}


// --------------------------------------------------------------------------
// Parser::EndElement(char *pElementName)
// --------------------------------------------------------------------------
void Parser::EndElement(char *pElementName)
{
    // Ignore GCC-XML element
    if (String::Equal(pElementName, "GCC_XML")) {
        return;
    }

    // "pop" the current element off of the element "stack"
    ref<Element> pElement = vCurrentElementsM.Last();

    // there had better have been one there
    if (pElement == NULL) {
        this->XmlError("Mismatched XML elements");
    }

    vCurrentElementsM.RemoveLast();
}


// --------------------------------------------------------------------------
// Parser::XmlError(const char *pFormat, ...) const
// --------------------------------------------------------------------------
void Parser::XmlError(const char *pFormat, ...) const
{
    va_list valist;

    va_start(valist, pFormat);

    Log::ErrorF("Line %lu: ", this->GetCurrentParserLineNumber());

    Log::ErrorV(pFormat, valist);

    va_end(valist);

    Log::Error("\n");

    Util::Terminate();
}


// --------------------------------------------------------------------------
// Parser::GetCurrentParserLineNumber() const
// --------------------------------------------------------------------------
U32 Parser::GetCurrentParserLineNumber() const
{
    return (U32) XML_GetCurrentLineNumber(xmlParserM);
}


// --------------------------------------------------------------------------
// Parser::Attribute::Attribute(char *pName, char *pValue)
// --------------------------------------------------------------------------
Parser::Attribute::Attribute(char *pName, char *pValue)
    : nameM(pName), valueM(pValue)
{
}

// --------------------------------------------------------------------------
// Parser::Element::Element(const Parser &parser, char **pAttributes)
// --------------------------------------------------------------------------
Parser::Element::Element(const Parser &parser, char *pElementName,
                         char **pAttributes)
    : elementNameM(pElementName)
{
    // Get current line number
    lineNumberM = parser.GetCurrentParserLineNumber();
    // Create attributes
    while (*pAttributes) {
        char *pAttributeName = *pAttributes++;
        char *pAttributeValue = *pAttributes++;
        // Skip attributes with no value
        if (!pAttributeValue || !*pAttributeValue) {
            continue;
        }
        ref<Parser::Attribute> pAttribute = 
            new Parser::Attribute(pAttributeName, pAttributeValue);
        attributesM.Append(pAttribute);
    }
}
