/*****************************************************************************\
 *                                                                           *
 * Configuration.cpp                                                         *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include <Essential/ProgramVariables.h>
#include <Essential/Log.h>
#include <Essential/Util.h>
#include "private/Configuration.h"

using namespace Essential;

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


static const char *usageMessageG = 
    "Usage: xrttigen [-D <definition>]... [-I <include_directory>]...\n"
    "                [-e <exclude_spec>]... [-h <header_file>]\n"
    "                [-i <include_spec>]... [-n] [-o <output_file>]\n"
    "                [-t <tmp file>] input_header_file...\n\n"
    "  -D:   Defines a preprocessor macro to be used when processing all "
    "input\n        header fles.\n"
    "  -I:   Specifies a directory to search for header files included by "
    "input\n        header files.\n"
    "  -e:   Gives a specification of a class or a set of classes to exclude "
    "from\n        generation.  Includes and excludes are evaluated in the "
    "order that\n        they appear on the command line, with later includes "
    "and excludes\n        taking precedence.\n"
    "  -h:   Adds a header file to the set of header files to include in the\n"
    "        generated output.  This is typically the same set of header "
    "files\n        used in the generation of the input gccxml output.\n"
    "  -i:   Gives a specification of a class or a set of classes to "
    "include in\n        generation.  Includes and excludes are evaluated in "
    "the order that\n        they appear on the command line, with later "
    "includes and excludes\n        taking precedence.\n"
    "  -n:   If present disables C++ RTTI support in the generated Xrtti "
    "code\n        this flag would be specified if your C++ code is "
    "being built without\n        C++ rtti support.\n"
    "  -o:   Names the output file to write the generated source to.  A "
    "value of\n        dash (-) indicates stdout.  Default is stdout.\n"
    "  -t:   Names the temporary file which should be used to store gccxml "
    "xml\n        output; this file will be created and then deleted during "
    "the run of\n        xrttigen.  Defaults to gccxml.out in the current "
    "directory.\n\n"
    "  input_header_file: Specifies a header files to process.\n\n"
    "  All names used with the -e and -i arguments may use simple "
    "wildcarding\n"
    "  in which either the first or last character of the name is an "
    "asterisk.\n\n";


static void UsageExit(bool stdout)
{
    if (stdout) {
        Log::Message(usageMessageG);
    }
    else {
        Log::Warning(usageMessageG);
    }

    Util::Terminate();
}


static inline bool IsIdentifierBeginning(const char &uc)
{
    return (((uc >= 'a') && (uc <= 'z')) || ((uc >= 'A') && (uc <= 'Z')) ||
            (uc == '_'));
}


static inline bool IsIdentifierPostBeginning(const char &uc)
{
    return (IsIdentifierBeginning(uc) || ((uc >= '0') && (uc <= '9')));
}


Configuration::Configuration(int argc, char **argv)
{
    ProgramVariables vars(argc, argv);

    if (vars.GetBool("-?", "HELP") || vars.GetBool("--help", "HELP")) {
        UsageExit(true);
    }

    // -D
    while (vars.HasVariable("-D", "DEFINITION")) {
        String definition = vars.GetString("-D", "DEFINITION", "");
        vDefinitionsM.Append(definition);
    }

    // -I
    while (vars.HasVariable("-I", "INCLUDE")) {
        String include = vars.GetString("-I", "INCLUDE", "");
        vIncludesM.Append(include);
    }

    // -h
    while (vars.HasVariable("-h", "HEADER")) {
        String header = vars.GetString("-h", "HEADER", "");
        vHeadersM.Append(header);
    }
    
    // -n
    disableRttiM = vars.GetBool("-n", "NO_RTTI");

    // -o (defaults to stdout)
    const char *o = vars.GetString("-o", "OUTPUT", NULL);
    if (o == NULL) {
        o = "-";
    }
    outFileM = o;

    // -t (defaults to gccxml.out)
    const char *t = vars.GetString("-t", "TEMPFILE", NULL);
    if (t == NULL) {
        t = "gccxml.out";
    }
    tmpFileM = t;

    // Get as of yet unused variables
    argc = vars.GetUnusedVariablesCount();
    argv = vars.GetUnusedVariables();

    // Skip program name
    argv++, argc--;

    // -i, -e, input_header_file
    while (argc) {
        if (String::Equal(*argv, "-i")) {
            argc--, argv++;
            if (argc == 0) {
                Log::Error("\nERROR: Parameter -i is missing argument.\n\n");
                UsageExit(false);
            }
            ref<Clude> pClude = new Clude(Clude::TypeIn, *argv);
            vCludesM.Append(pClude);
        }
        else if (String::Equal(*argv, "-e")) {
            argc--, argv++;
            if (argc == 0) {
                Log::Error("\nERROR: Parameter -i is missing argumen.t\n\n");
                UsageExit(false);
            }
            ref<Clude> pClude = new Clude(Clude::TypeEx, *argv);
            vCludesM.Append(pClude);
        }
        else {
            String inputHeaderFile(*argv);
            vInputsM.Append(inputHeaderFile);
        }
        argv++, argc--;
    }
    
    if (!vInputsM.Count()) {
        Log::Error("\nERROR: No input header files specified (at least one is "
                   "required).\n\n");
        UsageExit(false);
    }
}


bool Configuration::ShouldInclude(const String &name) const
{
    // Walk the list of "cludes" backwards, since each later one takes
    // precedence over prior ones
    U32 cludeCount = vCludesM.Count();

    for (U32 i = cludeCount; i > 0; i--) {
        ref<Clude> pClude = vCludesM[i - 1];

        // See if it's in there
        if (!pClude->Matches(name)) {
            // Nope, go to the next clude
            continue;
        }

        // OK, it matches - what does that mean?
        return (pClude->type == Clude::TypeIn) ? true : false;
    }

    // Default is to include everything
    return true;
}


Configuration::Clude::Clude(Clude::Type theType, char *theName)
{
    type = theType;

    String rawName = theName;

    if (rawName.StartsWith("subof:")) {
        match = MatchSubOf;
        name = String(&(theName[String::Length("subof:")]));
    }
    else if (rawName.StartsWith("has:")) {
        match = MatchHas;
        name = String(&(theName[String::Length("has:")]));
    }
    else {
        match = MatchExact;
    }

    if (rawName.StartsWith("*")) {
        wildcardStartM = true;
        wildcardEndM = false;
        name = String(&(theName[1]));
    }
    else if (rawName.EndsWith("*")) {
        wildcardStartM = false;
        wildcardEndM = true;
        name = String(theName, 0, String::Length(theName) - 1);
    }
    else {
        wildcardStartM = wildcardEndM = false;
        name = rawName;
    }
}


bool Configuration::Clude::Matches(const String &otherName)
{
    if (wildcardStartM) {
        return otherName.EndsWith(name);
    }
    else if (wildcardEndM) {
        return otherName.StartsWith(name);
    }
    else {
        return otherName.Equals(name);
    }
}


}; // namespace Xrtti
