/*****************************************************************************\
 *                                                                           *
 * Generator.cpp                                                             *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/


#include "private/Generator.h"

using namespace Essential;

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


U32 GeneratorObject::nextNumberG = 1;


Generator::Generator(Configuration &config, const ContextSet &contextSet)
    : configM(config)
{
    // Build up the vector of starting point GeneratorContexts
    U32 count = contextSet.GetContextCount();
    for (U32 i = 0; i < count; i++) {
        const Context *pContext = contextSet.GetContext(i);
        if (config.ShouldInclude(pContext->GetFullName())) {
            (void) this->GetGeneratorContext(pContext);
        }
    }
}

Generator::~Generator()
{
    Hashtable__ValueEnumeration<const Context *, GeneratorContext *>
        generatorContexts(htGeneratorContextsM);

    while (generatorContexts.HasElements()) {
        delete generatorContexts.GetNextElement();
    }

    Hashtable__ValueEnumeration<const Type *, GeneratorType *>
        generatorTypes(htGeneratorTypesM);

    while (generatorTypes.HasElements()) {
        delete generatorTypes.GetNextElement();
    }
}


bool Generator::Generate()
{
    String outFile = configM.GetOutputFile();
    File *pFile;
    bool createdFile;

    if (outFile == "-") {
        pFile = &(File::Stdout);
        createdFile = false;
    }
    else {
        pFile = new File();
        createdFile = true;
        if (!pFile->Open(outFile, File::OpenTypeWriteTruncate)) {
            delete pFile;
            return false;
        }
    }

    // Emit header
    this->EmitHeader(*pFile);

    // Emit XrttiAccess
    this->EmitXrttiAccess(*pFile);

    // Emit XrttiACcess static definitions
    this->EmitXrttiAccessStaticDefinitions(*pFile);

    // Emit declarations
    this->EmitDeclarations(*pFile);

    // Emit definitions
    this->EmitDefinitions(*pFile);

    // Emit register
    this->EmitRegister(*pFile);

    if (createdFile) {
        pFile->Close();
        delete pFile;
    }

    return true;
}


GeneratorContext *Generator::GetGeneratorContext(const Context *pContext)
{
    GeneratorContext *pRet;

    if (!htGeneratorContextsM.Get(pContext, /* returns */ pRet)) {
        switch (pContext->GetType()) {
        case Context::Type_Class:
            pRet = new GeneratorClass(*this, (const Class &) *pContext);
            break;
        case Context::Type_Namespace:
            pRet = new GeneratorNamespace(*this, (const Namespace &) *pContext);
            break;
        case Context::Type_Struct:
            pRet = new GeneratorStruct(*this, (const Struct &) *pContext);
            break;
        default: // Context::Type_Union:
            pRet = new GeneratorUnion(*this, (const Union &) *pContext);
            break;
        }
    }

    return pRet;
}


GeneratorType *Generator::GetGeneratorType(const Type *pType)
{
    GeneratorType *pRet;

    if (!htGeneratorTypesM.Get(pType, /* returns */ pRet)) {
        switch (pType->GetBaseType()) {
        case Type::BaseType_Enumeration:
            pRet = new GeneratorTypeEnumeration
                (*this, (const TypeEnumeration &) *pType);
            break;
        case Type::BaseType_Function:
            pRet = new GeneratorTypeFunction
                (*this, (const TypeFunction &) *pType);
            break;
        case Type::BaseType_Structure:
            pRet = new GeneratorTypeStructure
                (*this, (const TypeStructure &) *pType);
            break;
        default: // basic types
            pRet = new GeneratorType(*this, *pType);
            break;
        }
    }

    return pRet;
}


GeneratorEnumeration *Generator::
GetGeneratorEnumeration(const Enumeration *pEnumeration)
{
    GeneratorEnumeration *pRet;

    if (!htGeneratorEnumerationsM.Get(pEnumeration, /* returns */ pRet)) {
        pRet = new GeneratorEnumeration(*this, *pEnumeration);
    }

    return pRet;
}


void Generator::RegisterGeneratorContext(const Context *pContext,
                                         GeneratorContext *pGeneratorContext)
{
    htGeneratorContextsM.Put(pContext, pGeneratorContext);
}


void Generator::RegisterGeneratorType(const Type *pType,
                                      GeneratorType *pGeneratorType)
{
    htGeneratorTypesM.Put(pType, pGeneratorType);
}


void Generator::
RegisterGeneratorEnumeration(const Enumeration *pEnumeration,
                             GeneratorEnumeration *pGeneratorEnumeration)
{
    htGeneratorEnumerationsM.Put(pEnumeration, pGeneratorEnumeration);
}


/* static */
void Generator::EmitBooleanArgument(OutputStream &out, bool val, bool comma)
{
    out.WriteFA("        %s%s", val ? "true" : "false", comma ? ",\n" : "");
}


/* static */
void Generator::EmitStringArgument(OutputStream &out, const char *str, 
                                   bool comma)
{
    out.WriteFA("        \"%s\"%s", str, comma ? ",\n" : "");
}


/* static */
void Generator::EmitU32Argument(OutputStream &out, U32 number, bool comma)
{
    out.WriteFA("        %lu%s", number, comma ? ",\n" : "");
}


/* static */
void Generator::EmitAccessTypeArgument(OutputStream &out,
                                       AccessType accessType, bool comma)
{
    const char *str;

    switch (accessType) {
    case AccessType_Public:
        str = "Xrtti::AccessType_Public";
        break;
    case AccessType_Protected:
        str = "Xrtti::AccessType_Protected";
        break;
    default:
        str = "Xrtti::AccessType_Private";
        break;
    }

    out.WriteFA("        %s%s", str, comma ? ",\n" : "");
}


/* static */
void Generator::EmitArrayArgument(OutputStream &out, U32 count,
                                  const char *name, U32 number, bool comma)
{
    Generator::EmitU32Argument(out, count, true);

    if (count) {
        out.WriteFA("        _%lu_%s%s", number, name, comma ? ",\n" : "");
    }
    else {
        Generator::EmitU32Argument(out, 0, comma);
    }
}


/* static */
bool Generator::IsAccessible(const Context *pContext)
{
    while (true) {
        // Namespaces are always accessible
        if (pContext->GetType() == Context::Type_Namespace) {
            return true;
        }

        const Structure *pStructure = (const Structure *) pContext;

        // If it's anonymous or incomplete, then no it is not
        if (pStructure->IsAnonymous() || pStructure->IsIncomplete()) {
            return false;
        }

        // If it's not public, and it's not a friend of XrttiAccess, then
        // it's not accessible
        if ((pStructure->GetAccessType() != AccessType_Public) &&
            !IsXrttiAccessFriend(pStructure)) {
            return false;
        }

        // Check its context; we know there must be one because every
        // context terminates at the :: namespace
        pContext = pStructure->GetContext();
    }
}


/* static */
bool Generator::IsAccessible(const Member *pMember)
{
    // First off, the containing context must be accessible
    const Context &context = pMember->GetContext();
    if (!IsAccessible(&context)) {
        return false;
    }

    // Next up, if the containing context is a friend of XrttiAccess then
    // everything in it is accessible, so this member is
    if ((context.GetType() != Context::Type_Namespace) &&
        IsXrttiAccessFriend((const Structure *) &context)) {
        return true;
    }

    // Finally, it's only accessible if it's public
    return (pMember->GetAccessType() == AccessType_Public);
}


/* static */
bool Generator::IsPod(const Context *pContext)
{
    if (pContext->GetType() == Context::Type_Namespace) {
        return true;
    }

    const Structure *pStructure = (const Structure *) pContext;

    U32 count = pStructure->GetFieldCount();
    for (U32 i = 0; i < count; i++) {
        const Field &field = pStructure->GetField(i);

        // Ignore static fields; they aren't really in the structure
        if (field.IsStatic()) {
            continue;
        }

        // A POD class cannot have:
        // - protected or private fields
        if ((field.GetAccessType() == AccessType_Protected) ||
            (field.GetAccessType() == AccessType_Private)) {
            return false;
        }
        
        // A POD class cannot have:
        // - non-POD members
        const Xrtti::Type &type = field.GetType();

        // If it's a structure type
        if (type.GetBaseType() == Xrtti::Type::BaseType_Structure) {
            const Xrtti::TypeStructure &typeStructure =
                (const Xrtti::TypeStructure &) type;
            bool isPointer = false;
            U32 count = typeStructure.GetArrayOrPointerCount();
            for (U32 i = 0; i < count; i++) {
                if (typeStructure.GetArrayOrPointer(i).GetType() == 
                    ArrayOrPointer::Type_Pointer) {
                    isPointer = true;
                    break;
                }
            }
            
            // If it's not a pointer, and thus is embedded, then the
            // embedded structure must be POD
            if (!isPointer && 
                !Generator::IsPod(&typeStructure.GetStructure())) {
                return false;
            }
        }
        
        // A POD class cannot have:
        // - pointers to member methods
        // Not sure how to detect this
        
        // A POD class cannot have:
        // - references
        if (type.IsReference()) {
            return false;
        }
    }
    
    // A POD class cannot have:
    // - constructors
    if (pStructure->GetConstructorCount()) {
        return false;
    }
    
    // A POD class cannot have:
    // - base classes
    if (pStructure->GetBaseCount()) {
        return false;
    }

    if (pStructure->GetType() == Context::Type_Union) {
        return true;
    }

    const Struct *pStruct = (const Struct *) pStructure;

    // A POD class cannot have:
    // - virtual methods
    // - overloaded assignment operator
    count = pStruct->GetMethodCount();
    for (U32 i = 0; i < count; i++) {
        const Method &method = pStruct->GetMethod(i);

        if (method.IsVirtual()) {
            return false;
        }

        if (method.IsOperatorMethod() &&
            String::Equal(method.GetName(), "=")) {
            return false;
        }
    }

    return true;
}


/* static */
String Generator::GetTypeName(const Context *pContext)
{
    switch (pContext->GetType()) {
    case Context::Type_Class:
        if (((const Structure *) pContext)->HasStructureName()) {
            return (String("class ") + pContext->GetFullName());
        }
        else {
            return pContext->GetFullName();
        }
    case Context::Type_Namespace:
        return pContext->GetFullName();
    case Context::Type_Struct:
        if (((const Structure *) pContext)->HasStructureName()) {
            return (String("struct ") + pContext->GetFullName());
        }
        else {
            return pContext->GetFullName();
        }
    default: // Context::Type_Union
        if (((const Structure *) pContext)->HasStructureName()) {
            return (String("union ") + pContext->GetFullName());
        }
        else {
            return pContext->GetFullName();
        }
    }
}


/* static */
bool Generator::IsXrttiAccessFriend(const Structure *pStructure)
{
    U32 count = pStructure->GetFriendCount();
    for (U32 i = 0; i < count; i++) {
        if (String::Equal
            (pStructure->GetFriend(i).GetFullName(), "XrttiAccess")) {
            return true;
        }
    }

    return false;
}


void Generator::EmitHeader(OutputStream &out)
{
    // Emit the disclaimer
    out.WriteA("/****************************************************"
               "*************************\\\n"
               " *                                                         "
               "                  *\n"
               " *    AUTOGENERATED FILE - DO NOT EDIT - "
               "AUTOGENERATED FILE - DO NOT EDIT    *\n"
               " *                                                         "
               "                  *\n"
               " *      This file was generated by the "
               "xrttigen program.                     *\n"
               " *      Do not edit this file directly.  Edit the source "
               "file(s) instead     *\n"
               " *      and re-generate this file.                         "
               "                  *\n"
               " *                                                          "
               "                 *\n"
               " *    AUTOGENERATED FILE - DO NOT EDIT - "
               "AUTOGENERATED FILE - DO NOT EDIT    *\n"
               " *                                                          "
               "                 *\n"
               "\\************************************************************"
               "*****************/\n\n");

    // Emit the "includes"
    out.WriteA("#include <stddef.h>\n");
    out.WriteA("#include <typeinfo>\n");
    out.WriteA("#include <Xrtti/XrttiPrivate.h>\n");
    U32 count = configM.GetHeaderCount();
    for (U32 i = 0; i < count; i++) {
        String header = configM.GetHeader(i);
        if (header.StartsWith("<")) {
            out.WriteFA("#include %S\n", (const unichar *) header);
        }
        else {
            out.WriteFA("#include \"%S\"\n", (const unichar *) header);
        }
    }

    out.WriteA("\n");
}


void Generator::EmitXrttiAccess(OutputStream &out)
{
    out.WriteA("class XrttiAccess\n{\npublic:\n\n");

    this->EmitTypedefs(out);

    out.WriteA("\n");

    Hashtable__ValueEnumeration<const Context *, GeneratorContext *>
        generatorContexts(htGeneratorContextsM);

    while (generatorContexts.HasElements()) {
        generatorContexts.GetNextElement()->EmitXrttiAccess(out);
    }

    out.WriteA("};\n\n");
}


void Generator::EmitXrttiAccessStaticDefinitions(OutputStream &out)
{
    Hashtable__ValueEnumeration<const Context *, GeneratorContext *>
        generatorContexts(htGeneratorContextsM);

    while (generatorContexts.HasElements()) {
        generatorContexts.GetNextElement()->
            EmitXrttiAccessStaticDefinitions(out);
    }

    out.WriteA("\n");
}


void Generator::EmitTypedefs(OutputStream &out)
{
    Hashtable__ValueEnumeration<const Context *, GeneratorContext *>
        generatorContexts(htGeneratorContextsM);

    while (generatorContexts.HasElements()) {
        generatorContexts.GetNextElement()->EmitTypedefs(out);
    }
}


void Generator::EmitDeclarations(OutputStream &out)
{
    out.WriteA("namespace\n{\n\nstruct _\n{\n");

    Hashtable__ValueEnumeration<const Context *, GeneratorContext *>
        generatorContexts(htGeneratorContextsM);

    while (generatorContexts.HasElements()) {
        generatorContexts.GetNextElement()->EmitDeclaration(out);
    }

    Hashtable__ValueEnumeration<const Type *, GeneratorType *>
        generatorTypes(htGeneratorTypesM);

    while (generatorTypes.HasElements()) {
        generatorTypes.GetNextElement()->EmitDeclaration(out);
    }

    Hashtable__ValueEnumeration<const Enumeration *, GeneratorEnumeration *>
        generatorEnumerations(htGeneratorEnumerationsM);

    while (generatorEnumerations.HasElements()) {
        generatorEnumerations.GetNextElement()->EmitDeclaration(out);
    }

    out.WriteA("};\n\n}\n\n");
}


void Generator::EmitDefinitions(OutputStream &out)
{
    out.WriteA("namespace\n{\n");

    Hashtable__ValueEnumeration<const Context *, GeneratorContext *>
        generatorContexts(htGeneratorContextsM);

    while (generatorContexts.HasElements()) {
        generatorContexts.GetNextElement()->EmitDefinition(out);
    }

    Hashtable__ValueEnumeration<const Type *, GeneratorType *>
        generatorTypes(htGeneratorTypesM);

    while (generatorTypes.HasElements()) {
        generatorTypes.GetNextElement()->EmitDefinition(out);
    }

    Hashtable__ValueEnumeration<const Enumeration *, GeneratorEnumeration *>
        generatorEnumerations(htGeneratorEnumerationsM);

    while (generatorEnumerations.HasElements()) {
        generatorEnumerations.GetNextElement()->EmitDefinition(out);
    }

    out.WriteA("}\n\n");
}


void Generator::EmitRegister(OutputStream &out)
{
    out.WriteA("namespace\n{\n");

    if (htGeneratorContextsM.Count()) {
        out.WriteA("    static Xrtti::Context *contexts[] =\n    {\n");

        Hashtable__ValueEnumeration<const Context *, GeneratorContext *>
            generatorContexts(htGeneratorContextsM);

        while (generatorContexts.HasElements()) {
            GeneratorContext *pNext = generatorContexts.GetNextElement();
            out.WriteFA("        &_::_%lu%s\n", pNext->GetNumber(),
                        generatorContexts.HasElements() ? "," : "");
        }

        out.WriteA("    };\n\n");
    }

    out.WriteFA("    static Xrtti::CompiledRegister registration\n    (\n"
                "        %lu,\n        %s\n    );\n",
                htGeneratorContextsM.Count(),
                htGeneratorContextsM.Count() ? "contexts" : "0");

    out.WriteA("}\n");
}


}; // namespace Xrtti
