/*****************************************************************************\
 *                                                                           *
 * GeneratorField.cpp                                                        *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/


#include "private/Generator.h"

using namespace Essential;

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


GeneratorField::GeneratorField(Generator &generator, const Field &field)
    : GeneratorMember(generator, (const Member &) field), fieldM(field)
{
    pTypeM = generator.GetGeneratorType(&(field.GetType()));
}


void GeneratorField::EmitTypedefs(OutputStream &out)
{
    if (this->CanEmitGet() && fieldM.IsStatic() && fieldM.GetType().IsConst()) {
        pTypeM->EmitTypedef(out);
    }    
}


void GeneratorField::EmitXrttiAccess(OutputStream &out)
{
    if (this->CanEmitOffsetof()) {
        const Context &context = fieldM.GetContext();

        out.WriteFA("    static const Xrtti::U32 _%lu_offset = "
                    "offsetof(%S, %s);\n\n", this->GetNumber(), 
                    (const unichar *) Generator::GetTypeName(&context),
                    fieldM.GetName());
    }

    if (this->CanEmitGet()) {
        this->EmitGet(out);
    }
}

void GeneratorField::EmitXrttiAccessStaticDefinitions(OutputStream &out)
{
    if (this->CanEmitGet() && fieldM.IsStatic() && fieldM.GetType().IsConst()) {
        out.WriteFA("XrttiAccess::_%lu_type XrttiAccess::_%lu_faked = "
                    "%s::%s;\n\n", pTypeM->GetNumber(), this->GetNumber(), 
                    fieldM.GetContext().GetFullName(), fieldM.GetName());
    }
}


void GeneratorField::EmitDefinitionArguments(OutputStream &out)
{
    this->GeneratorMember::EmitDefinitionArguments(out);
    out.WriteA(",\n");
    
    out.WriteFA("        &_::_%lu,\n", pTypeM->GetNumber());

    Generator::EmitU32Argument(out, fieldM.GetBitfieldBitCount(), true);
    
    if (this->CanEmitOffsetof()) {
        Generator::EmitBooleanArgument(out, true, true);
        out.WriteFA("        XrttiAccess::_%lu_offset,\n", this->GetNumber());
    }
    else {
        Generator::EmitBooleanArgument(out, false, true);
        Generator::EmitU32Argument(out, 0, true);
    }

    // get
    if (this->CanEmitGet()) {
        out.WriteFA("        &XrttiAccess::_%lu_get", this->GetNumber());
    }
    else {
        Generator::EmitU32Argument(out, 0, false);
    }
}


bool GeneratorField::CanEmitOffsetof()
{
    if (String::Equal(fieldM.GetName(), "")) {
        return false;
    }

    if (!Generator::IsAccessible(&fieldM)) {
        return false;
        
    }

    const Context &context = fieldM.GetContext();
    
    return ((context.GetType() != Context::Type_Namespace) &&
            Generator::IsPod(&context));
}


bool GeneratorField::CanEmitGet()
{
    if (!Generator::IsAccessible(&fieldM)) {
        return false;
    }

    if (!fieldM.GetName()[0]) {
        return false;
    }

    if (fieldM.GetBitfieldBitCount()) {
        return false;
    }

    return true;
}


void GeneratorField::EmitGet(OutputStream &out)
{
    // First fake the field if it's a static const; this is the only way to
    // have an address to return (even though it won't be the address of
    // the actual static const)
    if (fieldM.IsStatic() && fieldM.GetType().IsConst()) {
        out.WriteFA("    static _%lu_type _%lu_faked;\n\n",
                    pTypeM->GetNumber(), this->GetNumber()); 
    }

    out.WriteFA("    static void *_%lu_get(void *pThis)\n"
                "    {\n", this->GetNumber());

    if (fieldM.IsStatic()) {
        if (fieldM.GetType().IsConst()) {
            out.WriteFA("        return (void *) & _%lu_faked;\n",
                        this->GetNumber());
        }
        else {
            out.WriteFA("        return (void *) & %s::%s;\n", 
                        fieldM.GetContext().GetFullName(), fieldM.GetName());
        }
    }
    else {
        out.WriteFA("        return (void *) & ((%S *) pThis)->%s;\n", 
                    (const unichar *) Generator::GetTypeName
                    (&(fieldM.GetContext())), fieldM.GetName());
    }
    
    out.WriteFA("    }\n\n");
}


}; // namespace Xrtti
