/*****************************************************************************\
 *                                                                           *
 * GeneratorMethod.cpp                                                       *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/


#include "private/Generator.h"

using namespace Essential;

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


GeneratorMethod::GeneratorMethod(Generator &generator, const Method &method)
    : GeneratorMember(generator, (const Member &) method), methodM(method),
      methodSignatureM(generator, method.GetSignature())
{
}


bool GeneratorMethod::CanInvoke()
{
    // Only non-anonymous, accessible context constructors can be emitted
    const Structure &structure = 
        (const Structure &) (methodM.GetContext());

    if (structure.IsAnonymous() || 
        (((structure.GetType() == Context::Type_Class) ||
          (structure.GetType() == Context::Type_Struct)) &&
         ((const Struct &) structure).IsAbstract())) {
        return false;
    }

    // Also, the method signature must be 

    return true;
}


void GeneratorMethod::EmitTypedefs(OutputStream &out)
{
    if (!Generator::IsAccessible(&methodM) || !this->CanInvoke()) {
        return;
    }

    methodSignatureM.EmitTypedefs(out);
}


void GeneratorMethod::EmitXrttiAccess(OutputStream &out)
{
    if (!Generator::IsAccessible(&methodM) || !this->CanInvoke()) {
        return;
    }

    out.WriteFA("    static void _%lu_invoke(void *pThis, void *pRet, "
                "void **pArgs)\n    {\n", this->GetNumber());

    out.WriteA("        ");

    const Type &returnType = methodM.GetSignature().GetReturnType();
    if ((returnType.GetBaseType() != Type::BaseType_Void) ||
        returnType.GetArrayOrPointerCount()) {
        bool refReturned = methodM.GetSignature().GetReturnType().IsReference();
        out.WriteFA("* (_%lu_type *%s) pRet = %s", 
                    methodSignatureM.GetReturnTypeNumber(),
                    refReturned ? "*" : "", refReturned ? "& " : "");
    }

    out.WriteFA("((%S *) pThis)->%s%s(", 
                (const unichar *) Generator::GetTypeName
                (&(methodM.GetContext())),
                methodM.IsOperatorMethod() ? "operator " : "", 
                methodM.GetName());

    methodSignatureM.EmitInvokeArguments(out);
    
    out.WriteA(");\n    }\n\n");
}


void GeneratorMethod::EmitExtraDeclarations(OutputStream &out)
{
    methodSignatureM.EmitDeclaration(out);
}


void GeneratorMethod::EmitExtraDefinitions(OutputStream &out)
{
    methodSignatureM.EmitDefinition(out);

    U32 count = methodSignatureM.GetArgumentCount();
    if (count) {
        out.WriteFA("    static char *_%lu_names[] =\n    {\n", 
                    this->GetNumber());
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("        \"%s\"%s\n", methodM.GetArgumentName(i),
                        (i < (count - 1)) ? "," : "");
        }
        out.WriteA("\n    };\n\n");
    }
}


void GeneratorMethod::EmitDefinitionArguments(OutputStream &out)
{
    this->GeneratorMember::EmitDefinitionArguments(out);
    out.WriteA(",\n");

    Generator::EmitBooleanArgument(out, methodM.IsOperatorMethod(), true);

    Generator::EmitBooleanArgument(out, methodM.IsConst(), true);

    Generator::EmitBooleanArgument(out, methodM.IsVirtual(), true);

    Generator::EmitBooleanArgument(out, methodM.IsPureVirtual(), true);

    out.WriteFA("        &_::_%lu,\n", methodSignatureM.GetNumber());

    if (methodSignatureM.GetArgumentCount()) {
        out.WriteFA("        _%lu_names,\n", this->GetNumber());
    }
    else {
        out.WriteA("        0,\n");
    }

    if (Generator::IsAccessible(&methodM) && this->CanInvoke()) {
        out.WriteFA("        &XrttiAccess::_%lu_invoke", this->GetNumber());
    }
    else {
        Generator::EmitU32Argument(out, 0, false);
    }
}


}; // namespace Xrtti
