/*****************************************************************************\
 *                                                                           *
 * Generator.cpp                                                             *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/


#include "private/Generator.h"

using namespace Essential;

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


GeneratorStructure::GeneratorStructure(Generator &generator, 
                                       const Structure &structure)
    : GeneratorContext(generator, (const Context &) structure), 
      rttiM(generator.GetConfiguration().GetRtti()), structureM(structure)
{
    U32 count = structure.GetBaseCount();

    if (count) {
        pBasesM = new GeneratorBase *[count];
        for (U32 i = 0; i < count; i++) {
            pBasesM[i] = new GeneratorBase
                (generator, structure.GetBase(i), this);
        }
    }
    else {
        pBasesM = 0;
    }

    count = structure.GetFriendCount();

    if (count) {
        pFriendsM = new GeneratorStructure *[count];
        for (U32 i = 0; i < count; i++) {
            pFriendsM[i] = (GeneratorStructure *)
                generator.GetGeneratorContext(&(structure.GetFriend(i)));
        }
    }
    else {
        pFriendsM = 0;
    }

    count = structure.GetFieldCount();

    if (count) {
        pFieldsM = new GeneratorField *[count];
        for (U32 i = 0; i < count; i++) {
            pFieldsM[i] = new GeneratorField(generator, structure.GetField(i));
        }
    }
    else {
        pFieldsM = 0;
    }

    count = structure.GetConstructorCount();

    if (count) {
        pConstructorsM = new GeneratorConstructor *[count];
        for (U32 i = 0; i < count; i++) {
            pConstructorsM[i] = new GeneratorConstructor
                (generator, structure.GetConstructor(i));
        }
    }
    else {
        pConstructorsM = 0;
    }

    if (structure.HasDestructor()) {
        pDestructorM = new GeneratorDestructor
            (generator, structure.GetDestructor());
    }
    else {
        pDestructorM = 0;
    }
}


GeneratorStructure::~GeneratorStructure()
{
    U32 count = structureM.GetBaseCount();
    for (U32 i = 0; i < count; i++) {
        delete pBasesM[i];
    }
    delete [] pBasesM;

    delete [] pFriendsM;

    count = structureM.GetFieldCount();
    for (U32 i = 0; i < count; i++) {
        delete pFieldsM[i];
    }
    delete [] pFieldsM;

    count = structureM.GetConstructorCount();
    for (U32 i = 0; i < count; i++) {
        delete pConstructorsM[i];
    }
    delete [] pConstructorsM;

    delete pDestructorM;
}


void GeneratorStructure::EmitXrttiAccess(OutputStream &out)
{
    U32 count = structureM.GetBaseCount();
    for (U32 i = 0; i < count; i++) {
        pBasesM[i]->EmitXrttiAccess(out);
    }

    count = structureM.GetFieldCount();
    for (U32 i = 0; i < count; i++) {
        pFieldsM[i]->EmitXrttiAccess(out);
    }

    // While emitting xrtti access for constructors and destructors, make
    // sure to note if the default constructor and/or default destructor
    // are inaccessible, which will prevent them from being emitted.  Also,
    // shortcircuit this process if the structure itself is not accessible.
    bool defaultConstructorAccessible, defaultDestructorAccessible;
    defaultConstructorAccessible = defaultDestructorAccessible =
        Generator::IsAccessible(&structureM);

    count = structureM.GetConstructorCount();
    for (U32 i = 0; i < count; i++) {
        GeneratorConstructor *pConstructor = pConstructorsM[i];
        pConstructor->EmitXrttiAccess(out);

        // If it's the default constructor, is it accessible?
        if (pConstructor->IsDefault() && !pConstructor->IsAccessible()) {
            defaultConstructorAccessible = false;
        }
    }

    if (pDestructorM) {
        pDestructorM->EmitXrttiAccess(out);
        if (!pDestructorM->IsAccessible()) {
            defaultDestructorAccessible = false;
        }
    }

    // Finally, only emit constructors and destructors if this structure
    // is not anonymous, abstract, or incomplete
    if (structureM.IsAnonymous() || structureM.IsIncomplete() ||
        (((structureM.GetType() == Context::Type_Class) ||
          (structureM.GetType() == Context::Type_Struct)) &&
         ((const Struct &) structureM).IsAbstract())) {
        return;
    }

    if (this->CanInvokeDefaultConstructor()) {
        // create
        out.WriteFA("    static void *_%lu_create()\n    {\n", 
                    this->GetNumber());
        out.WriteFA("        return new %S();\n    }\n\n", 
                    (const unichar *) Generator::GetTypeName(&structureM));

        // createarray
        out.WriteFA("    static void *_%lu_create_array(Xrtti::U32 count)\n"
                    "    {\n", this->GetNumber());
        out.WriteFA("        return new %S[count];\n    }\n\n", 
                    (const unichar *) Generator::GetTypeName(&structureM));
    }

    if (this->CanInvokeDefaultDestructor()) {
        // delete
        out.WriteFA("    static void _%lu_delete(void *pThis)\n    {\n",
                    this->GetNumber());
        out.WriteFA("        delete (%S *) pThis;\n    }\n\n", 
                    (const unichar *) Generator::GetTypeName(&structureM));

        // deletearray
        out.WriteFA("    static void _%lu_delete_array(void *pThis)\n    {\n",
                    this->GetNumber());
        out.WriteFA("        delete [] (%S *) pThis;\n    }\n\n", 
                    (const unichar *) Generator::GetTypeName(&structureM));
    }
}


void GeneratorStructure::EmitXrttiAccessStaticDefinitions(OutputStream &out)
{
    U32 count = structureM.GetFieldCount();
    for (U32 i = 0; i < count; i++) {
        pFieldsM[i]->EmitXrttiAccessStaticDefinitions(out);
    }
}


void GeneratorStructure::EmitNeededTypedefs(OutputStream &out)
{
    U32 count = structureM.GetFieldCount();
    for (U32 i = 0; i < count; i++) {
        pFieldsM[i]->EmitTypedefs(out);
    }

    count = structureM.GetConstructorCount();
    for (U32 i = 0; i < count; i++) {
        pConstructorsM[i]->EmitTypedefs(out);
    }
}

bool GeneratorStructure::CanInvokeDefaultConstructor()
{
    if (!this->CanInvokeDefaultDestructor()) {
        return false;
    }

    bool hasNonaccessibleConstructor = false;

    U32 count = structureM.GetConstructorCount();
    for (U32 i = 0; i < count; i++) {
        GeneratorConstructor *pConstructor = pConstructorsM[i];

        // If it's the default constructor, can it be invoked?
        if (pConstructor->IsDefault()) {
            return (pConstructor->IsAccessible() && pConstructor->CanInvoke());
        }
        else if (!pConstructor->IsAccessible()) {
            hasNonaccessibleConstructor = true;
        }
    }

    return hasNonaccessibleConstructor;
}


bool GeneratorStructure::CanInvokeDefaultDestructor()
{
    if (!Generator::IsAccessible(&structureM)) {
        return false;
    }

    if (pDestructorM) {
        return (pDestructorM->IsAccessible() && pDestructorM->CanInvoke());
    }

    return true;
}


void GeneratorStructure::EmitExtraDeclarations(OutputStream &out)
{
    U32 count = structureM.GetBaseCount();
    for (U32 i = 0; i < count; i++) {
        pBasesM[i]->EmitDeclaration(out);
    }

    count = structureM.GetFieldCount();
    for (U32 i = 0; i < count; i++) {
        pFieldsM[i]->EmitDeclaration(out);
    }
    
    count = structureM.GetConstructorCount();
    for (U32 i = 0; i < count; i++) {
        pConstructorsM[i]->EmitDeclaration(out);
    }

    if (pDestructorM) {
        pDestructorM->EmitDeclaration(out);
    }
}


void GeneratorStructure::EmitExtraDefinitions(OutputStream &out)
{
    U32 count = structureM.GetBaseCount();
    if (count) {
        for (U32 i = 0; i < count; i++) {
            pBasesM[i]->EmitDefinition(out);
        }
        out.WriteFA("    static Xrtti::Base *_%lu_bases[] =\n    {\n",
                    this->GetNumber());
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("        &_::_%lu%s\n", pBasesM[i]->GetNumber(),
                        (i < (count - 1)) ? "," : "");
        }
        out.WriteA("    };\n\n");
    }

    count = structureM.GetFriendCount();
    if (count) {
        out.WriteFA("    static Xrtti::Structure *_%lu_friends[] =\n    {\n",
                    this->GetNumber());
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("        &_::_%lu%s\n", pFriendsM[i]->GetNumber(),
                        (i < (count - 1)) ? "," : "");
        }
        out.WriteA("    };\n\n");
    }

    count = structureM.GetFieldCount();
    if (count) {
        for (U32 i = 0; i < count; i++) {
            pFieldsM[i]->EmitDefinition(out);
        }
        out.WriteFA("    static Xrtti::Field *_%lu_fields[] =\n    {\n",
                    this->GetNumber());
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("        &_::_%lu%s\n", pFieldsM[i]->GetNumber(),
                        (i < (count - 1)) ? "," : "");
        }
        out.WriteA("    };\n\n");
    }

    count = structureM.GetConstructorCount();
    if (count) {
        for (U32 i = 0; i < count; i++) {
            pConstructorsM[i]->EmitDefinition(out);
        }
        out.WriteFA("    static Xrtti::Constructor *"
                    "_%lu_constructors[] =\n    {\n", this->GetNumber());
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("        &_::_%lu%s\n", pConstructorsM[i]->GetNumber(),
                        (i < (count - 1)) ? "," : "");
        }
        out.WriteA("    };\n\n");
    }

    if (pDestructorM) {
        pDestructorM->EmitDefinition(out);
    }
}


void GeneratorStructure::EmitDefinitionArguments(OutputStream &out)
{
    this->GeneratorContext::EmitDefinitionArguments(out);
    out.WriteA(",\n");

    // bool isIncomplete
    Generator::EmitBooleanArgument(out, structureM.IsIncomplete(), true);

    // bool HasStructureName
    Generator::EmitBooleanArgument(out, structureM.HasStructureName(), true);
    
    // AccessType accessType
    Generator::EmitAccessTypeArgument(out, structureM.GetAccessType(), true);

    // std::type_info *pTypeInfo
    if (rttiM && !structureM.IsAnonymous() && !structureM.IsIncomplete() &&
        Generator::IsAccessible(&structureM)) {
        out.WriteFA("        &typeid(%s),\n", structureM.GetFullName());
    }
    else {
        out.WriteA("        0,\n");
    }

    // U32 baseCount
    // CompiledBase **pBases
    Generator::EmitArrayArgument
        (out, structureM.GetBaseCount(), "bases", this->GetNumber(), true);

    // U32 friendCount
    // CompiledStructure **pFriends
    Generator::EmitArrayArgument
        (out, structureM.GetFriendCount(), "friends", this->GetNumber(), true);
    
    // U32 fieldCount
    // CompiledField **pFields
    Generator::EmitArrayArgument
        (out, structureM.GetFieldCount(), "fields", this->GetNumber(), true);

    // bool isAnonymous
    Generator::EmitBooleanArgument(out, structureM.IsAnonymous(), true);

    // U32 constructorCount
    // CompiledConstructor **pConstructors
    Generator::EmitArrayArgument
        (out, structureM.GetConstructorCount(), "constructors",
         this->GetNumber(), true);

    // CompiledDestructor *pDestructor
    if (pDestructorM) {
        out.WriteFA("        &_%lu,\n", pDestructorM->GetNumber());
    }
    else {
        out.WriteFA("        0,\n");
    }

    // CompiledCreate *pCreate
    // CompiledCreateArray *pCreateArray
    if (this->CanInvokeDefaultConstructor()) {
        out.WriteFA("        &XrttiAccess::_%lu_create,\n", this->GetNumber());
        out.WriteFA("        &XrttiAccess::_%lu_create_array,\n", 
                    this->GetNumber());
    }
    else {
        Generator::EmitU32Argument(out, 0, true);
        Generator::EmitU32Argument(out, 0, true);
    }

    // CompiledDelete *pDelete
    // CompiledDeleteArray *pDeleteArray    
    if (this->CanInvokeDefaultDestructor()) {
        out.WriteFA("        &XrttiAccess::_%lu_delete,\n", this->GetNumber());
        out.WriteFA("        &XrttiAccess::_%lu_delete_array",
                    this->GetNumber());
    }
    else {
        Generator::EmitU32Argument(out, 0, true);
        Generator::EmitU32Argument(out, 0, false);
    }
}


}; // namespace Xrtti
