/*****************************************************************************\
 *                                                                           *
 * GeneratorType.cpp                                                         *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/


#include "private/Generator.h"

using namespace Essential;

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif


GeneratorType::GeneratorType(Generator &generator, const Type &type)
    : typedefEmittedM(false), typeM(type)
{
    generator.RegisterGeneratorType(&type, this);

    U32 count = type.GetArrayOrPointerCount();
    if (count) {
        pArrayOrPointersM = new GeneratorArrayOrPointer *[count];
        for (U32 i = 0; i < count; i++) {
            const ArrayOrPointer &arrayOrPointer = type.GetArrayOrPointer(i);
            if (arrayOrPointer.GetType() == ArrayOrPointer::Type_Array) {
                pArrayOrPointersM[i] = new GeneratorArray
                    ((const Array &) arrayOrPointer);
            }
            else {
                pArrayOrPointersM[i] = new GeneratorPointer
                    ((const Pointer &) arrayOrPointer);
            }
        }
    }
    else {
        pArrayOrPointersM = 0;
    }
}


GeneratorType::~GeneratorType()
{
    U32 count = typeM.GetArrayOrPointerCount();
    for (U32 i = 0; i < count; i++) {
        delete pArrayOrPointersM[i];
    }
    delete [] pArrayOrPointersM;
}


const char *GeneratorType::GetBaseTypeName()
{
    switch (typeM.GetBaseType()) {
    case Type::BaseType_Void:
        return "Type::BaseType_Void";
    case Type::BaseType_Bool:
        return "Type::BaseType_Bool";
    case Type::BaseType_Char:
        return "Type::BaseType_Char";
    case Type::BaseType_Unsigned_Char:
        return "Type::BaseType_Unsigned_Char";
    case Type::BaseType_WChar:
        return "Type::BaseType_WChar";
    case Type::BaseType_Short:
        return "Type::BaseType_Short";
    case Type::BaseType_Unsigned_Short:
        return "Type::BaseType_Unsigned_Short";
    case Type::BaseType_Int:
        return "Type::BaseType_Int";
    case Type::BaseType_Unsigned_Int:
        return "Type::BaseType_Unsigned_Int";
    case Type::BaseType_Long:
        return "Type::BaseType_Long";
    case Type::BaseType_Unsigned_Long:
        return "Type::BaseType_Unsigned_Long";
    case Type::BaseType_Long_Long:
        return "Type::BaseType_Long_Long";
    case Type::BaseType_Unsigned_Long_Long:
        return "Type::BaseType_Unsigned_Long_Long";
    case Type::BaseType_Float:
        return "Type::BaseType_Float";
    case Type::BaseType_Double:
        return "Type::BaseType_Double";
    case Type::BaseType_Long_Double:
        return "Type::BaseType_Long_Double";
    case Type::BaseType_Enumeration:
        return "Type::BaseType_Enumeration";
    case Type::BaseType_Function:
        return "Type::BaseType_Function";
    default: // BaseType_Structure
        return "Type::BaseType_Structure";
    }
}

// parts of a typedef:
// 
//      typedef type_name type_definition (function_arguments) ;
void GeneratorType::EmitTypedef(OutputStream &out)
{
    // If typedef has already been emitted, then don't emit it
    if (typedefEmittedM) {
        return;
    }

    typedefEmittedM = true;

    this->EmitPrerequisiteTypedefs(out);

    // typedef type_name
    out.WriteA("    typedef ");

    if (typeM.IsConst() && (typeM.GetArrayOrPointerCount() || 
                            typeM.IsReference())) {
        out.WriteA("const ");
    }

    if (typeM.IsVolatile()) {
        out.WriteA("volatile ");
    }

    out.WriteA(this->GetTypedefTypeName());

    // type_definition
    String tdef = String::Format("_%lu_type", this->GetNumber());

    // Don't include reference information; this is because we're just
    // typedef'ing enough for return type and argument type purposes, and
    // reference information is not included in that

    U32 count = typeM.GetArrayOrPointerCount();
    for (U32 i = 0; i < count; i++) {
        const ArrayOrPointer &arrayOrPointer = typeM.GetArrayOrPointer(i);
        if (arrayOrPointer.GetType() == ArrayOrPointer::Type_Array) {
            const Array &array = (const Array &) arrayOrPointer;
            if (array.IsUnbounded()) {
                tdef += "[]";
            }
            else {
                tdef += String::Format("[%lu]", array.GetElementCount());
            }
        }
        else {
            const Pointer &pointer = (const Pointer &) arrayOrPointer;
            if (pointer.IsVolatile()) {
                if (pointer.IsConst()) {
                    tdef -= " const volatile ";
                }
                else {
                    tdef -= " volatile ";
                }
            }
            else if (pointer.IsConst()) {
                tdef -= " const ";
            }
            tdef -= "*";
        }
        tdef -= "(";
        tdef += ")";
    }

    out.WriteA(" ");

    out.WriteA(tdef);

    // function_arguments
    this->EmitTypedefFunctionArguments(out);

    out.WriteA(";\n\n");
}


String GeneratorType::GetTypedefTypeName()
{
    // We know that this is a basic type, not a function, enumeration, or
    // structure
    switch (typeM.GetBaseType()) {
    case Type::BaseType_Void:
        return "void";
    case Type::BaseType_Bool:
        return "bool";
    case Type::BaseType_Char:
        return "char";
    case Type::BaseType_Unsigned_Char:
        return "unsigned char";
    case Type::BaseType_WChar:
        return "wchar_t";
    case Type::BaseType_Short:
        return "short";
    case Type::BaseType_Unsigned_Short:
        return "unsigned short";
    case Type::BaseType_Int:
        return "int";
    case Type::BaseType_Unsigned_Int:
        return "unsigned int";
    case Type::BaseType_Long:
        return "long";
    case Type::BaseType_Unsigned_Long:
        return "unsigned long";
    case Type::BaseType_Long_Long:
        return "long long";
    case Type::BaseType_Unsigned_Long_Long:
        return "unsigned long long";
    case Type::BaseType_Float:
        return "float";
    case Type::BaseType_Double:
        return "double";
    default: // Type::BaseType_Long_Double:
        return "long double";
    }
}


void GeneratorType::EmitExtraDeclarations(OutputStream &out)
{
    U32 count = typeM.GetArrayOrPointerCount();
    for (U32 i = 0; i < count; i++) {
        pArrayOrPointersM[i]->EmitDeclaration(out);
    }
}


void GeneratorType::EmitExtraDefinitions(OutputStream &out)
{
    U32 count = typeM.GetArrayOrPointerCount();
    if (count) {
        for (U32 i = 0; i < count; i++) {
            pArrayOrPointersM[i]->EmitDefinition(out);
        }
        out.WriteFA("    static Xrtti::ArrayOrPointer *"
                    "_%lu_array_or_pointers[] =\n    {\n", this->GetNumber());
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("        &_::_%lu%s\n", 
                        pArrayOrPointersM[i]->GetNumber(),
                        (i < (count - 1)) ? "," : "");
        }
        out.WriteFA("    };\n\n");
    }
}


void GeneratorType::EmitDefinitionArguments(OutputStream &out)
{
    out.WriteFA("        Xrtti::%s,\n", this->GetBaseTypeName());
    
    Generator::EmitBooleanArgument(out, typeM.IsConst(), true);

    Generator::EmitBooleanArgument(out, typeM.IsVolatile(), true);

    Generator::EmitBooleanArgument(out, typeM.IsReference(), true);

    Generator::EmitArrayArgument
        (out, typeM.GetArrayOrPointerCount(), "array_or_pointers", 
         this->GetNumber(), false);
}


String GeneratorType::GetCppTypeName()
{
    switch (typeM.GetBaseType()) {
    case Type::BaseType_Void:
        return "void";
    case Type::BaseType_Bool:
        return "bool";
    case Type::BaseType_Char:
        return "char";
    case Type::BaseType_Unsigned_Char:
        return "unsigned char";
    case Type::BaseType_WChar:
        return "wchar_t";
    case Type::BaseType_Short:
        return "short";
    case Type::BaseType_Unsigned_Short:
        return "unsigned short";
    case Type::BaseType_Int:
        return "int";
    case Type::BaseType_Unsigned_Int:
        return "unsigned int";
    case Type::BaseType_Long:
        return "long";
    case Type::BaseType_Unsigned_Long:
        return "unsigned long";
    case Type::BaseType_Long_Long:
        return "long long";
    case Type::BaseType_Unsigned_Long_Long:
        return "unsigned long long";
    case Type::BaseType_Float:
        return "float";
    case Type::BaseType_Double:
        return "double";
    case Type::BaseType_Long_Double:
        return "long double";
    case Type::BaseType_Enumeration:
        {
            const Enumeration &enumeration = 
                ((const TypeEnumeration &) typeM).GetEnumeration();
            return (String(enumeration.GetContext().GetFullName()) + 
                    "::" + enumeration.GetName());
        }
    case Type::BaseType_Function:
        return "todo";
    default: // BaseType_Structure
        return ((const TypeStructure &) typeM).GetStructure().GetFullName();
    }
}


}; // namespace Xrtti
